#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Interfaces describing the events and fields this package uses.

Also utility functions.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import traceback

from zope.deprecation import deprecated

from zope.schema import Text
from zope.schema import TextLine
from zope.interface import Interface
from zope.interface import Attribute
from zope.interface import providedBy
from zope.interface import implementer
from zope.schema import interfaces as sch_interfaces
try:
    from zope.schema._bootstrapfields import BeforeObjectAssignedEvent
except ImportError: # pragma: no cover
    # BWC for older zope.schema.
    from zope.schema._field import BeforeObjectAssignedEvent

__docformat__ = "restructuredtext en"

# pylint:disable=inherit-non-class,no-self-argument

class IBeforeSchemaFieldAssignedEvent(Interface):
    """
    An event sent when certain schema fields will be assigning an
    object to a property of another object.

    The interface
    :class:`zope.schema.interfaces.IBeforeObjectAssignedEvent` is a
    sub-interface of this one once this module is imported.
    """
    object = Attribute(u"The object that is going to be assigned. Subscribers may modify this")

    name = Attribute(u"The name of the attribute under which the object will be assigned.")

    context = Attribute(u"The context object where the object will be assigned to.")

# Make this a base of the zope interface so our handlers
# are compatible. This is dangerous if any lookups or registrations have already been done,
# as zope.interface maintains a cache of these things.
sch_interfaces.IBeforeObjectAssignedEvent.__bases__ = (IBeforeSchemaFieldAssignedEvent,)

@implementer(IBeforeSchemaFieldAssignedEvent)
class BeforeSchemaFieldAssignedEvent(object):

    def __init__(self, obj, name, context):
        self.object = obj
        self.name = name
        self.context = context

class IBeforeTextAssignedEvent(IBeforeSchemaFieldAssignedEvent):
    """
    Event for assigning text.
    """

    object = Text(title=u"The text being assigned.")

class IBeforeTextLineAssignedEvent(IBeforeTextAssignedEvent):  # ITextLine extends IText
    """
    Event for assigning text lines.
    """

    object = TextLine(title=u"The text being assigned.")

class IBeforeContainerAssignedEvent(IBeforeSchemaFieldAssignedEvent):
    """
    Event for assigning containers (__contains__).
    """

class IBeforeIterableAssignedEvent(IBeforeContainerAssignedEvent):
    """
    Event for assigning iterables.
    """

class IBeforeCollectionAssignedEvent(IBeforeIterableAssignedEvent):
    """
    Event for assigning collections.
    """

    object = Attribute(u"The collection being assigned. May or may not be mutable.")

class IBeforeSetAssignedEvent(IBeforeCollectionAssignedEvent):
    """
    Event for assigning sets.
    """

class IBeforeSequenceAssignedEvent(IBeforeCollectionAssignedEvent):
    """
    Event for assigning sequences.
    """

    object = Attribute(u"The sequence being assigned. May or may not be mutable.")

class IBeforeDictAssignedEvent(IBeforeIterableAssignedEvent):
    """
    Event for assigning dicts.
    """

# The hierarchy is IContainer > IIterable > ICollection > ISequence > [ITuple, IList]
# Also:         IContainer > IIterable > IDict
# Also:         IContainer > IIterable > ISet

@implementer(IBeforeTextAssignedEvent)
class BeforeTextAssignedEvent(BeforeSchemaFieldAssignedEvent):
    pass

@implementer(IBeforeTextLineAssignedEvent)
class BeforeTextLineAssignedEvent(BeforeTextAssignedEvent):
    pass


@implementer(IBeforeCollectionAssignedEvent)
class BeforeCollectionAssignedEvent(BeforeSchemaFieldAssignedEvent):
    object = None

@implementer(IBeforeSequenceAssignedEvent)
class BeforeSequenceAssignedEvent(BeforeCollectionAssignedEvent):
    pass

@implementer(IBeforeSetAssignedEvent)
class BeforeSetAssignedEvent(BeforeCollectionAssignedEvent):
    pass

@implementer(IBeforeDictAssignedEvent)
class BeforeDictAssignedEvent(BeforeSchemaFieldAssignedEvent):
    pass

BeforeObjectAssignedEvent = BeforeObjectAssignedEvent


# Set up the alias for InvalidValue. We need to use an alias so that
# try/except and subclass works as expected. The unfortunate side
# effect is that to preserve the constructor we need to swizzle the
# __init__ of sch_interfaces.InvalidValue (the swizzling is harmless;
# in zope.schema 4.7, InvalidValue inherits the Exception constructor,
# which forbids any kwargs but allows arbitrary *args; we keep that
# behaviour). This is deprecated and will be removed in the future.
InvalidValue = sch_interfaces.InvalidValue

def _InvalidValue__init__(self, *args, **kwargs):
    field = kwargs.pop('field', None)
    value = kwargs.pop('value', None)
    if kwargs:
        raise TypeError("Too many kwargs for function InvalidValue")
    # like normal
    sch_interfaces.ValidationError.__init__(self, *args)
    if field is not None or value is not None:
        self.with_field_and_value(field, value)


sch_interfaces.InvalidValue.__init__ = _InvalidValue__init__

del _InvalidValue__init__

deprecated('InvalidValue',
           "Use zope.schema.interfaces.InvalidValue.with_field_and_value.")

assert hasattr(sch_interfaces.InvalidValue, 'value')
assert hasattr(sch_interfaces.ValidationError, 'field')


class IFromObject(Interface):
    """
    Something that can convert one type of object to another,
    following validation rules (see :class:`zope.schema.interfaces.IFromUnicode`)
    """

    def fromObject(obj):
        """
        Attempt to convert the object to the required type following
        the rules of this object.  Raises a TypeError or
        :class:`zope.schema.interfaces.ValidationError` if this isn't
        possible.
        """

class IVariant(sch_interfaces.IField, IFromObject):
    """
    Similar to :class:`zope.schema.interfaces.IObject`, but
    representing one of several different types.

    If :meth:`fromObject` or :meth:`validate` fails, it should raise a
    :class:`VariantValidationError`.
    """

class VariantValidationError(sch_interfaces.ValidationError):
    """
    An error raised when a value is not suitable for any of the fields
    of the variant.

    The `errors` attribute is an ordered sequence of validation errors,
    with one raised by each field of the variant in turn.

    .. versionadded:: 1.8.0
    """

    #: A sequence of validation errors
    errors = ()

    def __init__(self, field, value, errors):
        super(VariantValidationError, self).__init__()
        self.with_field_and_value(field, value)
        self.errors = errors

    def _ex_details(self):
        lines = []
        for e in self.errors:
            info = traceback.format_exception_only(type(e), e)
            # Trim trailing newline
            info[-1] = info[-1].rstrip()
            lines.append('\n'.join(info))
        return lines

    def _with_details(self, opening, detail_formatter):
        lines = ['      ' + e for e in self._ex_details()]
        lines.append('    Field: ' + detail_formatter(self.field))
        lines.append('    Value: ' + detail_formatter(self.value))
        lines.append(opening)
        lines.reverse()
        return '\n'.join(lines)

    def __str__(self):
        s = super(VariantValidationError, self).__str__()
        return self._with_details(s, str)

    def __repr__(self):
        s = super(VariantValidationError, self).__repr__()
        return self._with_details(s, repr)

class IListOrTuple(sch_interfaces.IList):
    pass

def find_most_derived_interface(ext_self, iface_upper_bound, possibilities=None):
    """
    Search for the most derived version of the interface `iface_upper_bound`
    implemented by `ext_self` and return that. Always returns at least `iface_upper_bound`

    :keyword possibilities: An iterable of schemas to consider. If not given,
        all the interfaces provided by ``ext_self`` will be considered.
    """
    if possibilities is None:
        possibilities = providedBy(ext_self)
    _iface = iface_upper_bound
    for iface in possibilities:
        if iface.isOrExtends(_iface):
            _iface = iface
    return _iface

try:
    from dm.zope.schema.interfaces import ISchemaConfigured as _ISchemaConfigured
except ImportError:
    _ISchemaConfigured = Interface

class ISchemaConfigured(_ISchemaConfigured):
    """
    marker interface for ``SchemaConfigured`` classes.

    Used to facilitate the registration of forms and views.
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for eqhash.py

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import unittest


from ..eqhash import EqHash

from hamcrest import assert_that
from hamcrest import calling
from hamcrest import is_
from hamcrest import is_not
from hamcrest import raises

__docformat__ = "restructuredtext en"

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904



@EqHash('a', 'b')
class Thing(object):
    a = 'a'
    b = 'b'

@EqHash('a', 'b', superhash=True)
class Thing2(object):
    a = 'a'
    b = 'b'

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

@EqHash('a', 'b', include_type=True)
class NotThing(object):
    a = 'a'
    b = 'b'

@EqHash('c', include_super=True)
class ChildThing(Thing):
    c = 'c'

@EqHash('c', include_super=False)
class ChildThingNoSuper(Thing):
    c = 'c'

@EqHash(include_super=True)
class ChildThingNoNames(Thing):
    pass

@EqHash('a', 'b', 'c', 'd', 'e', 'f')
class ManyThing(object):
    a = 'a'
    b = 'b'
    c = 'c'
    d = 'd'
    e = 'e'
    f = 'f'

class TestEqHash(unittest.TestCase):

    def test_eq_hash(self):

        thing1 = Thing()
        thing2 = Thing()

        assert_that(thing1, is_(thing1)) # self
        assert_that(thing1, is_(thing2)) # commutative
        assert_that(thing2, is_(thing1))
        assert_that(hash(thing1), is_(hash(thing2)))

        assert_that(thing1, is_not(self)) # missing attr

        thing2.b = 'B'

        assert_that(thing1, is_not(thing2))
        assert_that(hash(thing1), is_not(hash(thing2)))

        # hamcrest calls == even for is_not
        assert thing1 != thing2
        assert thing1 != self # NotImplemented

    def test_eq_hash_mutates(self):

        thing_nosuperhash = Thing()
        thing_superhash = Thing2() # initially, no superhashing required

        assert_that(hash(thing_nosuperhash), is_(hash(thing_superhash)))

        # Ok, mutate one of the attributes to require superhashing
        thing_superhash.a = [1, 2]
        assert_that(hash(thing_superhash), is_(hash(thing_superhash)))

        thing_superhash2 = Thing2()
        thing_superhash2.a = [1, 2]
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))
        # And again, to prove we don't change unnecessarily
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))
        # But we can mutate the other attribute and we will change
        # hash strategies
        thing_superhash.b = {}
        thing_superhash2.b = {}
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))

        # OK, now that we've mutated, mutate again to something that's
        # not a dictionary, but is hashable and isn't iterable.
        class AThing(object):
            pass
        athing = AThing()
        hash(athing)
        thing_superhash.b = athing
        thing_superhash2.b = athing
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))

        # One more mutation for coverage
        thing_superhash.a = athing
        thing_superhash2.a = athing
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))
        assert_that(hash(thing_superhash2), is_(hash(thing_superhash)))


    def test_eq_hash_classes(self):
        # Default doesn't include classes

        assert_that(Thing(), is_(Thing2()))
        assert_that(hash(Thing()), is_(hash(Thing2())))

        # We can ask for it, but note that the order of arguments
        # matters
        assert_that(Thing(), is_(NotThing()))
        assert_that(NotThing(), is_not(Thing()))

    def test_eq_hash_super(self):
        thing1 = ChildThing()
        thing2 = ChildThing()

        assert_that(thing1, is_(thing2))

        assert_that(hash(thing1), is_(hash(thing2)))

        thing1.a = 'A'
        assert_that(thing1, is_not(thing2))
        assert_that(hash(thing1), is_not(hash(thing2)))

        thing3 = ChildThingNoNames()
        thing4 = ChildThingNoNames()
        assert_that(thing3, is_(thing4))
        assert_that(hash(thing3), is_(hash(thing4)))

    def test_eq_hash_no_super(self):
        thing1 = ChildThingNoSuper()
        thing2 = ChildThingNoSuper()

        assert_that(thing1, is_(thing2))
        assert_that(hash(thing1), is_(hash(thing2)))

        thing1.a = 'A'
        assert_that(thing1, is_(thing2))
        assert_that(hash(thing1), is_(hash(thing2)))

        thing1.c = 'C'
        assert_that(thing1, is_not(thing2))
        assert_that(hash(thing1), is_not(hash(thing2)))

    def test_bad_construct(self):
        assert_that(calling(EqHash), raises(TypeError, "Asking to hash"))
        assert_that(calling(EqHash).with_args(foo=True),
                    raises(TypeError, "Unexpected keyword"))


class TestSuperHash(unittest.TestCase):

    def superhash(self, arg):
        # Use the old location to make sure it works
        from nti.schema import schema
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            return schema._superhash(arg) # pylint:disable=no-member

    def test_iterable(self):
        assert_that(hash(self.superhash([1, 3, 5])),
                    is_(hash(self.superhash([x for x in [1, 3, 5]]))))

        assert_that(self.superhash([1, 2]), is_not(self.superhash([2, 1])))
        assert_that(hash(self.superhash([1, 2])), is_not(hash(self.superhash([2, 1]))))

    def test_nested_dict(self):
        d = {1: 1,
             2: [1, 2, 3],
             3: {4: [4, 5, 6]}}

        assert_that(self.superhash(d),
                    is_(
                        ((1, 1),
                         (2, (1, 2, 3)),
                         (3, ((4, (4, 5, 6)),)))
                    ))

        assert_that(hash(self.superhash(d)),
                    is_(-6213620179105025536))


def test_suite():
    import doctest
    suite = unittest.defaultTestLoader.loadTestsFromName(__name__)
    suite.addTest(
        doctest.DocTestSuite('nti.schema.eqhash')
    )

    return suite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for fieldproperty.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import doctest
import unittest

from Acquisition import Implicit
from ExtensionClass import Base
from zope.interface import Interface
from zope.interface import implementer
from zope.schema.interfaces import SchemaNotProvided
from zope.schema.interfaces import WrongType

from nti.schema.field import Object
from nti.schema.field import ValidTextLine as TextLine
from nti.schema.fieldproperty import AcquisitionFieldProperty
from nti.schema.fieldproperty import AdaptingFieldProperty
from nti.schema.fieldproperty import AdaptingFieldPropertyStoredThroughField
from nti.schema.fieldproperty import UnicodeConvertingFieldProperty
from nti.schema.fieldproperty import createDirectFieldProperties
from nti.schema.fieldproperty import createFieldProperties
from nti.testing.matchers import aq_inContextOf

from hamcrest import assert_that
from hamcrest import calling
from hamcrest import has_key
from hamcrest import has_length
from hamcrest import has_property
from hamcrest import is_
from hamcrest import is_not
from hamcrest import none
from hamcrest import raises

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904,inherit-non-class

__docformat__ = "restructuredtext en"

does_not = is_not


class TestAqFieldProperty(unittest.TestCase):

    def test_aq_property(self):

        class IBaz(Interface):
            pass

        class IFoo(Interface):
            ob = Object(IBaz)

        @implementer(IBaz)
        class Baz(object):
            pass

        class BazAQ(Implicit, Baz):
            pass

        @implementer(IFoo)
        class Foo(Base):
            ob = AcquisitionFieldProperty(IFoo['ob'])

        assert_that(Foo, has_property('ob', is_(AcquisitionFieldProperty)))

        # pylint:disable=blacklisted-name
        foo = Foo()
        assert_that(foo, has_property('ob', none()))

        foo.ob = Baz()
        assert_that(foo, has_property('ob', is_not(aq_inContextOf(foo))))

        foo.ob = BazAQ()
        assert_that(foo, has_property('ob', aq_inContextOf(foo)))

class TestCreateFieldProperties(unittest.TestCase):

    def test_create_field_properties(self):
        class IBaz(Interface):
            pass

        class IA(Interface):
            a = TextLine(title=u"a")

        class IB(IA):
            b = Object(IBaz)

        class A(object):
            createFieldProperties(IA)

        class B(object):
            createDirectFieldProperties(IB, adapting=True)

        assert_that(A.__dict__, has_key('a'))
        assert_that(B.__dict__, has_key('b'))
        assert_that(B.__dict__, does_not(has_key('a')))
        assert_that(B.__dict__['b'], is_(AdaptingFieldProperty))

        # And nothing extra crept in, just the four standard things
        # __dict__, __doct__, __module__, __weakref__, and b
        assert_that(B.__dict__, has_length(5))

class TestUnicodeFieldProperty(unittest.TestCase):

    def test_set_bytes(self):
        class IA(Interface):
            a = TextLine(title=u"a")

        class A(object):
            a = UnicodeConvertingFieldProperty(IA['a'])

        a = A()
        a.a = b'abc'

        assert_that(a, has_property('a', is_(u'abc')))

class TestAdaptingFieldProperty(unittest.TestCase):

    def test_(self):
        class IBaz(Interface):
            pass

        class IFoo(Interface):
            ob = Object(IBaz)

        @implementer(IBaz)
        class Baz(object):
            pass

        class CantConform(object):
            pass

        class Conforms(object):
            def __conform__(self, iface):
                return Baz()

        for fp in AdaptingFieldProperty, AdaptingFieldPropertyStoredThroughField:

            assert_that(calling(fp).with_args(None), raises(WrongType))

            @implementer(IFoo)
            class O(object):
                ob = fp(IFoo['ob'])

            obj = O()

            # First, can't set it
            with self.assertRaises(SchemaNotProvided):
                obj.ob = CantConform()

            # But this can be adapted
            obj.ob = Conforms()
            assert_that(obj.ob, is_(Baz))

def test_suite():
    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        doctest.DocTestSuite("nti.schema.fieldproperty"),
    ))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
nti.schema.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import zope.i18nmessageid

__docformat__ = "restructuredtext en"

MessageFactory = zope.i18nmessageid.MessageFactory('nti.dataserver')



# -*- coding: utf-8 -*-
"""
Test support for schemas and interfaces, mostly in the form of
Hamcrest matchers.

The following functions can still be imported from here, but have been
moved to :mod:`nti.testing.matchers`:

- :func:`~nti.testing.matchers.provides`
- :func:`~nti.testing.matchers.implements`
- :func:`~nti.testing.matchers.verifiably_provides`
- :func:`~nti.testing.matchers.validly_provides`
- :func:`~nti.testing.matchers.validated_by`
- :func:`~nti.testing.matchers.not_validated_by`
"""

from __future__ import division
from __future__ import print_function

from zope.deprecation import moved
moved('nti.testing.matchers')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
For producing a JSON schema appropriate for use by clients, based on a
Zope schema.

The ``TAG`` constants are intended to be set as (boolean) tagged values
on fields of interfaces, helping determine how the schema is built.

..  note:: This schema is ad-hoc and non-standard.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from copy import copy

try:
    from collections.abc import Sequence
except ImportError: # pragma: no cover
    # Python 2
    from collections import Sequence

from numbers import Number

from six import integer_types
from six import string_types
from six import text_type

from zope.i18n import translate

from zope.interface.interfaces import IMethod
from zope.interface.interfaces import IInterface

from zope.schema import interfaces as sch_interfaces
from zope.schema import vocabulary as sch_vocabulary

from nti.schema.interfaces import IVariant
from nti.schema.interfaces import find_most_derived_interface

__docformat__ = "restructuredtext en"


#: Don't display this by default in the UI
TAG_HIDDEN_IN_UI = "nti.dataserver.users.field_hidden_in_ui"

#: Qualifying details about how the field should be treated, such as data source
#: The ``UI_TYPE`` values are defined known values for this tag.
TAG_UI_TYPE = 'nti.dataserver.users.field_type'

#: Overrides the value from the field itself
TAG_REQUIRED_IN_UI = 'nti.dataserver.users.field_required'

#: Overrides the value from the field itself, if true
TAG_READONLY_IN_UI = 'nti.dataserver.users.field_readonly'

#: A tagged value that echos semi-arbitrary data back in the schema
#: under the key ``application_info``.
#:
#: The value for this field *must* be a mapping object whose keys
#: are strings and whose values are JSON serializable. Text values
#: (specifically `zope.i18nmessageid.Message` objects) will be translated
#: when put into the schema. The object in the schema *will not* be the object
#: in the tagged value, and constituent composite values will also be copied.
#: (In case any post-processing of the schema is done based on current user,
#: request or site, this isolates such modifications.)
#:
#: .. versionadded:: 1.12.0
TAG_APPLICATION_INFO = 'nti.schema.jsonschema.field_application_info'

# Known types
#: The email type
UI_TYPE_EMAIL = 'nti.dataserver.users.interfaces.EmailAddress'

#: An email type that is stored as a non-recoverable hash.
#: The value is chosen so that a begins-with test will match
#: either this or :const:`UI_TYPE_EMAIL`, making validation easier
UI_TYPE_HASHED_EMAIL = UI_TYPE_EMAIL + ":Hashed"

#: Something that can be set once, typically during account creation
UI_TYPE_ONE_TIME_CHOICE = 'nti.dataserver.users.interfaces.OneTimeChoice'

def get_ui_type_from_interface(iface):
    ui_type = iface.getName()
    ui_type = ui_type[1:] if ui_type.startswith('I') else ui_type
    return ui_type
_interface_to_ui_type = interface_to_ui_type = get_ui_type_from_interface # BWC

def get_ui_type_from_field_interface(field):
    derived_field_iface = find_most_derived_interface(field, sch_interfaces.IField)
    if derived_field_iface is not sch_interfaces.IField:
        ui_type = get_ui_type_from_interface(derived_field_iface)
        return ui_type
    return None
_ui_type_from_field_iface = ui_type_from_field_iface = get_ui_type_from_field_interface # BWC

def get_ui_types_from_field(field):
    ui_type = ui_base_type = None
    _type = getattr(field, '_type', None)
    if isinstance(_type, type):
        ui_type = _type.__name__
    elif isinstance(_type, tuple):
        # Most commonly lists subclasses. Most commonly lists subclasses of strings
        if all((issubclass(x, string_types) for x in _type)):
            ui_type = 'basestring'
        elif all((issubclass(x, integer_types) for x in _type)):
            ui_type = 'int'
        elif all((issubclass(x, float) for x in _type)):
            ui_type = 'float'
        elif all((issubclass(x, Sequence) for x in _type)):
            ui_type = 'list'
    else:
        ui_type = get_ui_type_from_field_interface(field)

    if ui_type in ('unicode', 'str', 'basestring'):
        # These are all 'string' type

        # Can we be more specific?
        ui_type = get_ui_type_from_field_interface(field)
        if ui_type and ui_type not in ('TextLine', 'Text'): # pragma: no cover
            # Yes we can
            ui_base_type = 'string'
        else:
            ui_type = 'string'
            ui_base_type = 'string'
    elif ui_type in ('Sequence', 'MutableSequence'):
        ui_type = 'list'
    elif ui_type in ('Mapping', 'MutableMapping'):
        ui_type = 'dict'
    elif ui_type in ('Number', 'float', 'Decimal', 'Complex', 'Real', 'Rational'):
        ui_base_type = 'float'
    elif ui_type in ('int', 'long', 'Integral'):
        ui_base_type = 'int'
    elif ui_type in ('bool',):
        ui_base_type = 'bool'
    return ui_type, ui_base_type

_ui_type_from_field = ui_type_from_field = get_ui_types_from_field # BWC

def get_data_from_choice_field(v, base_type=None):
    # Vocabulary could be a name or the vocabulary itself
    choices = ()
    vocabulary = None
    if sch_interfaces.IVocabulary.providedBy(v.vocabulary): # pragma: no cover
        vocabulary = v.vocabulary
    elif isinstance(v.vocabularyName, string_types):
        name = v.vocabularyName
        vocabulary = sch_vocabulary.getVocabularyRegistry().get(None, name)

    if vocabulary is not None:
        choices = []
        tokens = []
        for term in vocabulary:
            # For BWC, we do different things depending on whether
            # there is a title or not
            if getattr(term, 'title', None):
                try:
                    # like nti.externalization, but without the dependency
                    choice = term.toExternalObject()
                except AttributeError: # pragma: no cover
                    choice = {
                        'token': term.token,
                        'value': term.value,
                        'title': term.title
                    }

                choices.append(choice)
            else: # pragma: no cover
                choices.append(term.token)  # bare; ideally this would go away
            tokens.append(term.token)

        # common case, these will all be the same type
        if  not base_type \
            and all((isinstance(x, string_types) for x in tokens)):
            base_type = 'string'
    return choices, base_type
_process_choice_field = process_choice_field = get_data_from_choice_field


class JsonSchemafier(object):

    def __init__(self, schema, readonly_override=None, context=None):
        """
        Create a new schemafier.

        :param schema: The zope schema (interface) to use.
        :param bool readonly_override: If given, a boolean value that will replace all
            readonly values in the schema.
        :param context: The context passed to :func:`zope.i18n.translate`
        """
        self.schema = schema
        self.readonly_override = readonly_override
        self.context = context

    def _iter_names_and_descriptions(self):
        """
        Return an iterable across the names and descriptions of the schema.

        Subclass hook to change what is considered.
        """
        return self.schema.namesAndDescriptions(all=True)

    def allow_field(self, name, field):
        """
        Return if the field is allowed in the external schema.

        By default, this checks to see if the field has a true value for the
        tag `TAG_HIDDEN_IN_UI`, or if the field's name starts with an underscore.

        .. versionchanged:: 1.12.0
           This will now be called for nested fields, such as the ``value_type`` or
           ``key_type`` of collections and mappings. In those cases, commonly the name will
           be the empty string.
        """
        if field.queryTaggedValue(TAG_HIDDEN_IN_UI) or name.startswith('_'):
            return False
        return True

    def get_ui_types_from_field(self, field):
        """
        Return the type and base type for the specified field
        """
        return get_ui_types_from_field(field)
    ui_types_from_field = get_ui_types_from_field # BWC

    def get_data_from_choice_field(self, field, base_type=None):
        """
        Return the choices and base type for the specified field
        """
        return get_data_from_choice_field(field, base_type)
    process_choice_field = get_data_from_choice_field # BWC

    def post_process_field(self, name, field, item_schema):
        pass

    def bind(self, schema):
        clone = self.__class__.__new__(self.__class__)
        clone.__dict__.update(self.__dict__)
        clone.schema = schema
        return clone

    def make_schema(self):
        """
        Create the JSON schema.

        Individual fields of the schema will be checked and returned. See the various
        ``TAG`` constants for ways that the schema externalization can be influenced.

        :return: A dictionary consisting of dictionaries, one for each field. All the keys
            are strings and the values are strings, bools, numbers, or lists of primitives.
            Will be suitable for writing to JSON.
        """
        ext_schema = {}
        for k, v in self._iter_names_and_descriptions():
            __traceback_info__ = k, v
            if IMethod.providedBy(v):
                continue
            # v could be a schema field or an interface.Attribute
            if not self.allow_field(k, v):
                # If we disallow at the top level, we don't even
                # hint to its existence.
                continue

            item_schema = self._make_field_schema(v, k)

            self.post_process_field(k, v, item_schema)
            ext_schema[k] = item_schema

        return ext_schema

    def _translate(self, text):
        return translate(text, context=self.context)

    def _make_field_schema(self, field, name=None):
        name = name or field.__name__ or ''
        if not self.allow_field(name, field):
            # Disallowed, but we're already expecting a return value to put
            # in a dictionary.
            return None

        required = getattr(field, 'required', None)
        required = field.queryTaggedValue(TAG_REQUIRED_IN_UI) or required
        readonly_override = self.readonly_override
        if readonly_override is not None:
            readonly = readonly_override
        else:
            readonly = getattr(field, 'readonly', False)
            readonly = field.queryTaggedValue(TAG_READONLY_IN_UI) or readonly

        ui_base_type = None
        item_schema = {
            'name': name,
            'required': required,
            'readonly': readonly,
        }

        ui_type = field.queryTaggedValue(TAG_UI_TYPE)
        if not ui_type:
            ui_type, ui_base_type = self.get_ui_types_from_field(field)
        else:
            _, ui_base_type = self.get_ui_types_from_field(field)

        item_schema['type'] = ui_type
        item_schema['base_type'] = ui_base_type

        # Now add supplemental information about nested data.

        # Note that we're consistent to always provide *something*
        # for these optional keys, if they are expected to be there based on the
        # type of the field, it just may be None if it was excluded (or empty,
        # if all variants were excluded).

        if sch_interfaces.IChoice.providedBy(field):
            choices, base_type = self.get_data_from_choice_field(field, ui_base_type)
            item_schema['choices'] = choices
            item_schema['base_type'] = base_type

        if IVariant.providedBy(field):
            # 'fields' is not actually declared in the IVariant
            # interface; that's ok, we couldn't handle it automatically anyway.
            item_schema['value_type_options'] = [
                self._make_field_schema(the_field)
                for the_field
                in field.fields
                if self.allow_field(the_field.__name__, the_field)
            ]

        application_info = field.queryTaggedValue(TAG_APPLICATION_INFO) or {}
        item_schema['application_info'] = {
            k: self._translate(v) if isinstance(v, text_type) else copy(v)
            for k, v in application_info.items()
        }

        self._fill_field_values(item_schema, field)

        return item_schema

    def _fill_field_values(self, item_schema, field,
                           # Certain things from IField that we exclude
                           # because they are not useful externally or may
                           # expose sensitive information. `validate_invariants`
                           # is from the Object field.
                           # TODO: What if we tagged these values on import with TAG_HIDDEN_IN_UI
                           # and checked for that here?
                           _excluded=('missing_value', 'default', 'validate_invariants'),
                           # For safety, we'll refuse to output a value we don't understand.
                           # We expect dicts to have come from us, creating schemas.
                           _allowed_value_types=string_types + (Number, type(None), dict)):
        derived_field_iface = find_most_derived_interface(field, sch_interfaces.IField)
        if not derived_field_iface or derived_field_iface is sch_interfaces.IField:
            return

        # Take anything it has as a primitive, query it, and put it in the schema.
        for name, field_field in derived_field_iface.namesAndDescriptions(all=True):
            if name in item_schema or name in _excluded:
                # Cheap test first: if we've already got it, ignore it
                continue
            if not sch_interfaces.IField.providedBy(field_field):
                continue

            # A few things we handle specially.
            value = field_field.query(field)
            if sch_interfaces.IField.providedBy(value):
                # It is another field, yay, like value_type or key_type,
                # so we need to recurse
                value = self._make_field_schema(value)
            elif IInterface.providedBy(value):
                schemafier = self.bind(value)
                value = schemafier.make_schema()
            elif isinstance(value, text_type):
                value = self._translate(value)

            if not isinstance(value, _allowed_value_types):
                continue # pragma: no cover
            item_schema[name] = value



#!/usr/bin/env python

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904


from hamcrest import is_not
from hamcrest import contains

from hamcrest import has_length
from hamcrest import assert_that
from hamcrest import has_property
from hamcrest import calling
from hamcrest import raises
does_not = is_not

import unittest

from zope import interface
from zope.component import eventtesting

from . import SchemaLayer
from . import IUnicode

from ..field import Number
from ..field import Object
from ..field import DictFromObject
from ..field import ValidTextLine as TextLine

from ..schema import PermissiveSchemaConfigured
from ..schema import SchemaConfigured


from ..interfaces import IBeforeDictAssignedEvent


class TestMisc(unittest.TestCase):

    def test_fixup_name(self):
        from zope.schema.fieldproperty import FieldPropertyStoredThroughField

        field = Object(IUnicode)
        field.__name__ = 'field'

        field_property = FieldPropertyStoredThroughField(field)
        field = field_property.field

        assert_that(field, has_property('__name__', '__st_field_st'))
        assert_that(field, has_property('__fixup_name__', 'field'))

class TestSchemaConfigured(unittest.TestCase):

    def test_init(self):
        class IA(interface.Interface):
            field = Number()

        @interface.implementer(IA)
        class A(PermissiveSchemaConfigured):
            pass

        # All of these are valid ways to call it
        A()
        A(thing='abc')
        a = A(field=1)
        assert_that(a, has_property('field', 1))

        # But we can flip off the extra
        A.SC_PERMISSIVE = False
        A()
        assert_that(calling(A).with_args(thing=1),
                    raises(TypeError, 'non schema keyword'))
        a = A(field=1)
        assert_that(a, has_property('field', 1))

    def test_readonly(self):
        from nti.schema.fieldproperty import createDirectFieldProperties
        class IA(interface.Interface):
            field = Number(readonly=True,
                           required=False,
                           default=1)

        @interface.implementer(IA)
        class A(SchemaConfigured):
            createDirectFieldProperties(IA)

        a = A()
        assert_that(a, has_property('field', 1))

    def test_property_raises_exception(self):
        # https://github.com/NextThought/nti.schema/issues/11
        class IA(interface.Interface):
            field = Number(readonly=True,
                           required=False,
                           default=1)

        @interface.implementer(IA)
        class A(SchemaConfigured):
            @property
            def field(self):
                raise ValueError("bad field")

        # We can still create it without raising an AttributeError;
        # in fact, the ValueError is propagated
        assert_that(calling(A), raises(ValueError, "bad field"))


class TestConfigured(unittest.TestCase):

    layer = SchemaLayer


    def test_dict(self):

        dict_field = DictFromObject(key_type=TextLine(), value_type=Number())
        dict_field.__name__ = 'dict'

        class X(object):
            pass

        x = X()
        dict_field.set(x, dict_field.fromObject({u'k': u'1'}))

        assert_that(x, has_property('dict', {u'k': 1.0}))

        events = eventtesting.getEvents(IBeforeDictAssignedEvent)
        assert_that(events, has_length(1))
        assert_that(events, contains(has_property('object', {'k': 1.0})))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helpers for hashing and equality based on a list of names.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import operator

import six

__docformat__ = "restructuredtext en"

def _superhash_force(value):
    # Called when we know that we can't hash the value.
    # Dict?
    try:
        # Sort these, they have no order
        items = sorted(value.items())
    except AttributeError:
        # mutable iterable, which we must not sort
        items = value

    return tuple([_superhash(item)
                  for item
                  in items])

def _superhash(value):
    """
    Returns something that's hashable, either the value itself,
    or a tuple that can in turn be hashed
    """
    try:
        # We used to think that by returning the original value, if it was hashable,
        # we may get better ultimate hash results;
        # the cost is hashing that value twice.
        # However, Python guarantees that the hash of an integer
        # *is* the integer (with a few exceptions), so we should be
        # just fine returning the hash value....except, Python 3 changes this
        # and the hash is less often the same. So we go back to hashing things twice.
        hash(value)
    except TypeError:
        return _superhash_force(value)
    else:
        return value

def EqHash(*names,
           **kwargs):
    """
    EqHash(*names, include_super=False, superhash=False, include_type=False)

    A class decorator factory for the common pattern of writing
    ``__eq__``/``__ne__`` and ``__hash__`` methods that check the same
    list of attributes on a given object.

    Right now, you must pass as individual arguments the property
    names to check; in the future, you may be able to pass a schema
    interface that defines the property names. Property names are compared
    for equality in the order they are given, so place the cheapest first.

    Additional parameters are only available via keywords::

      >>> @EqHash('a', 'b')
      ... class Thing(object):
      ...   a = 1
      ...   b = 2
      >>> hash(Thing()) == (hash(('a', 'b')) ^ hash((1, 2)))
      True

      >>> @EqHash('c', include_super=True)
      ... class ChildThing(Thing):
      ...   c = 3
      >>> hash(ChildThing()) != hash(Thing()) != 0
      True

    :keyword include_super: If set to ``True`` (*not* the default)
        then the equality (and perhaps hash) values of super will be considered.
    :keyword superhash: If set to ``True`` (*not* the default),
        then the hash function will be made to support certain
        mutable types (lists and dictionaries) that ordinarily cannot
        be hashed. Use this only when those items are functionally
        treated as immutable.
    :keyword include_type: If set to ``True`` (*not* the default),
        equality will only be true if the other object is an instance
        of the class this is declared on. Use this only when there are
        a series of subclasses who differ in no attributes but should not
        compare equal to each other. Note that this can lead to violating
        the commutative property.

    """

    _include_super = kwargs.pop('include_super', False)
    superhash = kwargs.pop("superhash", False)
    _include_type = kwargs.pop('include_type', False)

    if kwargs:
        raise TypeError("Unexpected keyword args", kwargs)
    if not names and not _include_super and not _include_type:
        raise TypeError("Asking to hash/eq nothing, but not including super or type")


    def x(cls):
        __eq__, __hash__, __ne__ = _eq_hash(cls, names,
                                            _include_super, _include_type, superhash)
        cls.__eq__ = __eq__
        cls.__hash__ = __hash__
        cls.__ne__ = __ne__
        return cls
    return x

def _make_eq(cls, names, include_super, include_type):
    # 1 and 0 are constants and faster to load than the globals True/False
    # (in python 2)

    eq_stmt = 'def __eq__(self, other'
    if include_type or include_super:
        # capture the type
        eq_stmt += ', cls=cls'
    eq_stmt += '):\n'
    eq_stmt += '    if self is other: return 1\n'
    if include_type:
        eq_stmt += '    if not isinstance(other, cls): return 0\n'
    if include_super:
        eq_stmt += '    s = super(cls, self).__eq__(other)\n'
        eq_stmt += '    if s is NotImplemented or not s: return s\n'

    # We take these one at a time (rather than using
    # operator.attrgetter). In the cases where some attributes
    # are computed, this can be more efficient if we discover
    # a mismatch early. Also, it lets us easily distinguish
    # between an AttributeError on self (which is a
    # programming error in calling EqHash) or the other object
    for name in names:
        eq_stmt += '    a = self.' + name + '\n'
        eq_stmt += '    try:\n        b = other.' + name + '\n'
        eq_stmt += '    except AttributeError: return NotImplemented\n'
        eq_stmt += '    if a != b: return 0\n\n'

    eq_stmt += '    return 1'

    # Must use a custom dictionary under Py3
    lcls = dict(locals())
    six.exec_(eq_stmt, globals(), lcls)

    return lcls['__eq__']

def _eq_hash(cls, names, include_super, include_type, superhash): # pylint:disable=I0011,W0622,R0912
    names = tuple((str(x) for x in names)) # make sure they're native strings, not unicode on Py2
    # We assume the class hierarchy of these objects does not change
    if include_super:
        superclass = cls.__mro__[1]
        superclass_hash = superclass.__hash__

    __eq__ = _make_eq(cls, names, include_super, include_type)

    def __ne__(self, other):
        eq = __eq__(self, other)
        if eq is NotImplemented:
            return eq
        return not eq

    # Our contract for include_super says that hashing
    # may or may not be included. It shouldn't affect the results
    # if we do not actually include it, unless there are no values
    # being hashed from this object. However, for consistency,
    # we always include it if asked
    seed = hash(names)
    if include_type:
        seed += hash(cls)

    if superhash:
        # We assume that instances that use superhash will have
        # roughly the same shape, and not all attributes will need to be
        # super-hashed. When an attribute does need to be super-hashed, it will
        # need to be super-hashed for all instances. Worst case scenario, this winds up
        # always using the superhash for all attributes of all instances, but if we're lucky
        # only a small number of the same attributes will need to be superhashed.

        class Transformers(list):
            mutated = False

        transformers = Transformers([None for _ in names])

        def _hash(values):
            # Hopefully in most cases everything is actually hashable.
            # This gets our overhead down to the lowest possible.
            if not transformers.mutated:
                try:
                    return hash(values)
                except TypeError:
                    pass

            # Ok, we found something that can't actually be hashed. Darn.
            # Replace every non-Hashable transformer with a call to superhash.
            transformers.mutated = True
            # Snap. Lets hope that we already checked on what needs to be superhashed
            # and if so we'll try that.
            try:
                return hash(tuple([transformer(value) if transformer is not None else value
                                   for transformer, value
                                   in zip(transformers, values)]))
            except TypeError:
                # Snap. Something changed.
                for i, value in enumerate(values):
                    if transformers[i] is _superhash:
                        # We've reached our limit. Nothing else to do
                        # for this one.
                        continue

                    if transformers[i] is _superhash_force:
                        try:
                            _superhash_force(value)
                        except TypeError:
                            # OK, this field alternates between
                            # being hashable and nat being hashable. Deal with that.
                            transformers[i] = _superhash

                    try:
                        # We could check isinstance(value, collections.Hashable), but
                        # this is slightly more general, albeit probably slower.
                        hash(value)
                    except TypeError:
                        transformers[i] = _superhash_force

            # Ok, good to go. Let's try it.
            return hash(tuple([transformer(value) if transformer is not None else value
                               for transformer, value
                               in zip(transformers, values)]))
    else:
        # No need to try to wrap in a tuple or anything, we can
        # just directly call the hash builtin. We'll get passed either
        # a tuple of values.
        _hash = hash

    # Unlike __eq__, we use operator.attrgetter because we're always
    # going to request all the names. In tests, this is ~30% faster than
    # a manual loop (for two to three names).
    if not names:
        # Well, ok, nothing special in this class. We can't pass that to attrgetter,
        # though, it needs at least one name. Make sure to return a tuple for
        # consistency.
        def attrgetter(_):
            return ()
    else:
        # This will return a tuple of the values of the names.
        attrgetter = operator.attrgetter(*names)

    def __hash__(self):
        h = seed
        if include_super:
            h ^= superclass_hash(self) << 2

        # If we or-equal for every attribute separately, we
        # easily run the risk of saturating the integer. So we callect
        # all attributes down to one tuple to hash
        h ^= _hash(attrgetter(self))
        return h

    return __eq__, __hash__, __ne__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Event handlers.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.component import adapter
from zope.component import handle

from nti.schema.interfaces import IBeforeSchemaFieldAssignedEvent

__docformat__ = "restructuredtext en"


@adapter(IBeforeSchemaFieldAssignedEvent)
def before_object_assigned_event_dispatcher(event):
    """
    Listens for :mod:`zope.schema` fields to fire
    :class:`~nti.schema.interfaces.IBeforeSchemaFieldAssignedEvent`, and re-dispatches
    these events based on the value being assigned, the object being
    assigned to, and of course the event (note that
    :class:`~zope.schema.interfaces.IBeforeObjectAssignedEvent` is a
    sub-interface of :class:`~nti.schema.interfaces.IBeforeSchemaFieldAssignedEvent`).

    This is analogous to
    :func:`zope.component.event.objectEventNotify`
    """
    handle(event.object, event.context, event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Schema fields.

These fields produce better validation errors than the standard
:mod:`zope.schema` fields do. All the standard fields are also aliased
to be imported from this module.

.. TODO: This module is big enough it should be factored into a package and sub-modules.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import numbers
import re

try:
    import collections.abc as abcs
except ImportError: # pragma: no cover
    # Python 2
    import collections as abcs

from six import string_types
from six import text_type
from zope import interface
from zope import schema
from zope.deferredimport import deprecatedFrom
from zope.event import notify
import zope.interface.common.idatetime
from zope.cachedescriptors.property import Lazy

from zope.schema import interfaces as sch_interfaces
from zope.schema.interfaces import IFromBytes
from zope.schema.interfaces import IFromUnicode

from zope.schema import Bool
from zope.schema import Choice
from zope.schema import Date
from zope.schema import Datetime
from zope.schema import Decimal
from zope.schema import Dict
from zope.schema import FrozenSet
from zope.schema import Iterable
from zope.schema import List
from zope.schema import Mapping
from zope.schema import MutableMapping
from zope.schema import MutableSequence
from zope.schema import Sequence
from zope.schema import Set
from zope.schema import Text
from zope.schema import TextLine
from zope.schema import Timedelta
from zope.schema import Tuple
from zope.schema import Object as _ObjectBase

from zope.schema import Complex
from zope.schema import Real
from zope.schema import Rational
from zope.schema import Integral



from nti.schema import MessageFactory as _
from nti.schema.interfaces import BeforeDictAssignedEvent
from nti.schema.interfaces import BeforeObjectAssignedEvent
from nti.schema.interfaces import BeforeSchemaFieldAssignedEvent
from nti.schema.interfaces import BeforeSequenceAssignedEvent
from nti.schema.interfaces import BeforeSetAssignedEvent
from nti.schema.interfaces import BeforeTextAssignedEvent
from nti.schema.interfaces import BeforeTextLineAssignedEvent
from nti.schema.interfaces import IFromObject
from nti.schema.interfaces import IListOrTuple
from nti.schema.interfaces import IVariant
from nti.schema.interfaces import VariantValidationError


__docformat__ = "restructuredtext en"

# Re-export some things as part of our public API so we can
# later re-implement them locally if needed
__all__ = [
    'Bool',
    'Choice',
    'Complex',
    'Date',
    'Datetime',
    'Decimal',
    'DecodingValidTextLine',
    'Dict',
    'DictFromObject',
    'FieldValidationMixin',
    'Float',
    'FrozenSet',
    'HTTPURL',
    'IndexedIterable',
    'Int',
    'Integral',
    'Iterable',
    'List',
    'ListOrTuple',
    'ListOrTupleFromObject',
    'Mapping',
    'MutableMapping',
    'MutableSequence',
    'Number',
    'Object',
    'ObjectLen',
    'Rational',
    'Real',
    'Set',
    'Sequence',
    'Text',
    'TextLine',
    'Timedelta',
    'Tuple',
    'TupleFromObject',
    'UniqueIterable',
    'ValidBytes',
    'ValidBytesLine',
    'ValidChoice',
    'ValidDatetime',
    'ValidRegEx',
    'ValidRegularExpression',
    'ValidSet',
    'ValidText',
    'ValidTextLine',
    'ValidURI',
    'Variant',
]

# BWC alias, not in __all__
DateTime = Datetime

deprecatedFrom(
    "Moved to nti.schema.schema",
    "nti.schema.schema",
    'SchemaConfigured',
)


def _do_set(self, context, value, cls, factory):
    try:
        event = factory(value, self.__name__, context)
        notify(event)
        value = event.object
        super(cls, self).set(context, value)
    except sch_interfaces.ValidationError as e: # pragma: no cover
        # This shouldn't happen, set() doesn't typically validate.
        self._reraise_validation_error(e, value) # pylint:disable=protected-access

def __make_set(cls, eventfactory):

    def set(self, context, value): # pylint:disable=redefined-builtin
        return _do_set(self, context, value, cls, eventfactory)

    cls.set = set

def __with_set(eventfactory=BeforeSchemaFieldAssignedEvent):
    def X(cls):
        __make_set(cls, eventfactory)
        return cls
    return X

def _fixup_Object_field(field):
    # If possible, make the field provide IFromObject
    if not IFromObject.providedBy(field):
        if isinstance(field, _ObjectBase):
            # An object field can do this with a tiny bit of help.
            def _object_fromObject(value):
                if not field.schema.providedBy(value) and value != field.missing_value:
                    # Allow the TypeError or LookupError to propagate,
                    # signalling nti.externalization that it doesn't need to
                    # try to adapt again.
                    value = field.schema(value)
                field.validate(value)
                return value
            field.fromObject = _object_fromObject
            interface.alsoProvides(field, IFromObject)
        elif sch_interfaces.ICollection.providedBy(field):
            # Has a value_type
            try:
                # pylint:disable=protected-access
                _SequenceFromObjectMixin._validate_contained_field(field.value_type)
            except sch_interfaces.SchemaNotProvided:
                # Nothing we can do
                pass
            else:
                def _collection_fromObject(value):
                    return _SequenceFromObject(field).fromObject(value)
                field.fromObject = _collection_fromObject
                interface.alsoProvides(field, IFromObject)
        elif sch_interfaces.IMapping.providedBy(field):
            # Has a value_type and key_type
            try:
                # pylint:disable=protected-access
                _SequenceFromObjectMixin._validate_contained_field(field.value_type)
                _SequenceFromObjectMixin._validate_contained_field(field.key_type)
            except sch_interfaces.SchemaNotProvided:
                # Nothing we can do
                pass
            else:
                def _map_fromObject(value):
                    return _MapFromObject(field).fromObject(value)
                field.fromObject = _map_fromObject
                interface.alsoProvides(field, IFromObject)
    return field


class FieldValidationMixin(object):
    """
    A field mixin that causes slightly better errors to be created.
    """

    @property
    def __fixup_name__(self):
        """
        The :class:`zope.schema.fieldproperty.FieldPropertyStoredThroughField` class mangles
        the field name; this undoes that mangling.

        .. seealso:: `nti.schema.fieldproperty.field_name`
        """
        from nti.schema.fieldproperty import field_name
        return field_name(self)

    def _fixup_validation_error_args(self, e, value):
        # Called when the exception has one argument, which is usually, though not always,
        # the message
        e.args = (value, e.args[0], self.__fixup_name__)

    def _fixup_validation_error_no_args(self, e, value):
        # Called when there are no arguments
        e.args = (value, str(e), self.__fixup_name__)

    def _reraise_validation_error(self, e, value, _raise=False):
        # This must be called inside a try/except with an active exception.
        assert e.field is not None, "A _validate method failed to use ex.with_field_and_value()"
        # As-of zope.schema 4.7.0, this is largely obsolete.
        # Arguments are provided; we stop calling this in many cases.
        if len(e.args) == 1:  # typically the message is the only thing
            self._fixup_validation_error_args(e, value)
        elif not e.args:
            self._fixup_validation_error_no_args(e, value)
        if _raise:
            raise e
        raise # pylint:disable=misplaced-bare-raise

    def _fixup_too_short(self, e, value):
        # Note we're capitalizing the field in the message.
        e.i18n_message = _(
            u'${field} is too short. Please use at least ${minLength} characters.',
            mapping={'field': self.__fixup_name__.capitalize(),
                     'minLength': e.bound}
        )
        e.args = (self.__fixup_name__.capitalize() + ' is too short.',
                  self.__fixup_name__,
                  value)

    def _fixup_too_long(self, e, value):
        e.i18n_message = _(
            u'${field} is too long. ${max_size} character limit.',
            mapping={'field': self.__fixup_name__.capitalize(),
                     'max_size': e.bound}
        )
        e.args = (self.__fixup_name__.capitalize() + ' is too long.',
                  self.__fixup_name__,
                  value)

    def _validate(self, value):
        try:
            super(FieldValidationMixin, self)._validate(value)
        except sch_interfaces.WrongType as e:
            assert e.expected_type is not None, "The expected_type should be provided"
            raise
        except sch_interfaces.WrongContainedType:
            raise
        except sch_interfaces.TooShort as e:
            self._fixup_too_short(e, value)
            raise
        except sch_interfaces.TooLong as e:
            self._fixup_too_long(e, value)
            raise
        except sch_interfaces.SchemaNotProvided as e:
            assert e.schema is not None, "The schema should be provided"
            raise
        except sch_interfaces.OutOfBounds as e:
            assert e.bound is not None, "The bound should be provided"
            raise
        except sch_interfaces.ValidationError as e:
            self._reraise_validation_error(e, value)


@interface.implementer(sch_interfaces.IObject)
class ValidDatetime(FieldValidationMixin, Datetime):
    """
    Unlike the standard datetime, this will check that the
    given object is an instance of IDatetime, and raise
    the same error as object does.
    """

    schema = zope.interface.common.idatetime.IDateTime

    def _validate(self, value):
        try:
            super(ValidDatetime, self)._validate(value)
        except sch_interfaces.WrongType as e:
            raise sch_interfaces.SchemaNotProvided(
                value, e.__doc__,
                self.__fixup_name__, self.schema,
                list(interface.providedBy(value))).with_field_and_value(self, value)

        # schema has to be provided by value
        if not self.schema.providedBy(value):  # pragma: no cover
            raise sch_interfaces.SchemaNotProvided().with_field_and_value(self, value)

def _iteritems(o):
    meth = getattr(o, 'iteritems', None) or getattr(o, 'items')
    return meth()

class Object(FieldValidationMixin, _ObjectBase):
    """
    Improved ``zope.schema.Object``.
    """

class _FieldConverter(object):

    def __init__(self, field):
        self.field = field

    @Lazy
    def fromObject(self):
        return getattr(self.field, 'fromObject', None)

    @Lazy
    def fromUnicode(self):
        return getattr(self.field, 'fromUnicode', None)

    @Lazy
    def fromBytes(self):
        return getattr(self.field, 'fromBytes', None)

    def __call__(self, value):
        # pylint:disable=too-many-function-args

        # Take make it easier on implementers of fromObject,
        # if they also implement fromBytes/fromUnicode, we call those
        # first.
        if isinstance(value, bytes) and self.fromBytes is not None:
            return self.fromBytes(value)

        if isinstance(value, text_type) and self.fromUnicode is not None:
            return self.fromUnicode(value)

        if self.fromObject is not None:
            return self.fromObject(value)

        # Last resort, see if it can be validated by the field as-is.
        # Let this raise whatever error it wants to, but be sure
        # validation errors have the details filled in.
        try:
            self.field.validate(value)
            return value
        except sch_interfaces.ValidationError as ex:
            if ex.field is None:
                ex.field = self.field
            if ex.value is None:
                ex.value = value
            raise


@interface.implementer(IVariant)
class Variant(FieldValidationMixin, schema.Field):
    """
    Similar to :class:`zope.schema.Object`, but accepts one of many
    non-overlapping interfaces.

    .. versionchanged:: 1.9.0

        Implementations of :class:`zope.schema.interfaces.ICollection`
        (like :class:`zope.schema.List`) and
        :class:`zope.schema.interfaces.IMapping` (like
        :class:`zope.schema.Dict`) will automatically be given a ``fromObject`` method
        when they are used as a field of this object *if* their *value_type* is an
        :class:`zope.schema.interfaces.IObject` (recursively).
    """

    fields = ()

    def __init__(self, fields, variant_raise_when_schema_provided=False, **kwargs):
        """
        :param fields: A list or tuple of field instances.
        :keyword variant_raise_when_schema_provided: If ``True``, then
            if a value is provided to ``validate`` that implements
            the schema of a particular field, and that field raised
            a validation error, that error will be propagated instead
            of the error raised by the last field, and no additional fields
            will be asked to do validation.

        """
        if not fields or not all((sch_interfaces.IField.providedBy(x) for x in fields)):
            raise sch_interfaces.SchemaNotProvided(sch_interfaces.IField)

        # assign our children first so anything we copy to them as a result of the super
        # constructor (__name__) gets set
        self.fields = [_fixup_Object_field(field) for field in fields]
        for f in self.fields:
            f.__parent__ = self

        self._raise_when_provided = variant_raise_when_schema_provided
        super(Variant, self).__init__(**kwargs)

    def __get_name(self):
        return self.__dict__.get('__name__', '')

    def __set_name(self, name):
        self.__dict__['__name__'] = name
        for field in self.fields:
            field.__name__ = name
    __name__ = property(__get_name, __set_name)

    def getExtraDocLines(self):
        lines = super(Variant, self).getExtraDocLines()
        lines.append(".. rubric:: Possible Values")

        for field in self.fields:
            lines.append(".. rubric:: Option")
            lines.append(field.getDoc())

        return lines

    def __repr__(self):
        return "<%s.%s at 0x%x name=%r interface=%r context=%r fields=%r>" % (
            type(self).__module__, type(self).__name__,
            id(self), self.__name__,
            self.interface, self.context, self.fields
        )

    def bind(self, obj):
        # The fields member really does exist
        # pylint:disable=no-member
        clone = super(Variant, self).bind(obj)
        clone.fields = [x.bind(obj) for x in clone.fields]
        for f in clone.fields:
            f.__parent__ = clone
        return clone

    def _validate(self, value):
        super(Variant, self)._validate(value)
        errors = []
        for field in self.fields:
            try:
                field.validate(value)
                # one of them accepted, yay!
                return
            except sch_interfaces.ValidationError as e:
                if (self._raise_when_provided
                        and hasattr(field, 'schema')
                        and field.schema.providedBy(value)):
                    self._reraise_validation_error(e, value)
                    raise AssertionError("This is never reached")
                else:
                    errors.append(e)
        try:
            raise VariantValidationError(self, value, errors)
        finally:
            # break cycles
            e = errors = None

    def fromObject(self, obj):
        """
        Similar to `fromUnicode`, attempts to turn the given object
        into something acceptable and valid for this field. Raises a
        `~.VariantValidationError` if this isn't possible. Adaptation
        is attempted in the order in which fields were given to the
        constructor. Some fields cannot be used to adapt.

        .. versionchanged:: 1.8.0
           Raise `~.VariantValidationError` instead of whatever
           last error we got.
        .. versionchanged:: 1.9.1
           Respect ``self.missing_value`` and don't raise an exception
           if it is passed to this method and we're not required.
        """
        if obj == self.missing_value:
            if self.required:
                raise sch_interfaces.RequiredMissing(self.__name__).with_field_and_value(self, obj)

            # If we're not required, we still attempt all the adaptation steps
            # just in case some field can convert us very nicely, but
            # we won't raise the final exception

        errors = []

        for field in self.fields:
            try:
                converter = _FieldConverter(field)

                # Try to convert and validate. This calls fromXXX
                # if defined, and otherwise validates. The fromXXX methods
                # are also supposed to validate, so validation should be done
                # when this returns.
                return converter(obj)
            except (TypeError, sch_interfaces.ValidationError) as ex:
                # Nope, no good
                errors.append(ex)

        # We get here if nothing worked.
        if obj == self.missing_value:
            # Well it's the missing value. By definition we're not required
            # at this point (we would have raised already), so we can just return it.
            assert not self.required
            return self.missing_value

        try:
            raise VariantValidationError(self, obj, errors)
        finally:
            # break cycles
            ex = errors = None

    _EVENT_TYPES = (
        (string_types, BeforeTextAssignedEvent),
        (abcs.Mapping, BeforeDictAssignedEvent),
        (abcs.Sequence, BeforeSequenceAssignedEvent),
        (object, BeforeObjectAssignedEvent)
    )

    def set(self, context, value):
        # Try to determine the most appropriate event to fire
        # Order matters. It would kind of be nice to direct this to the appropriate
        # field itself, but that's sort of hard.
        for kind, factory in self._EVENT_TYPES:
            if isinstance(value, kind):
                _do_set(self, context, value, Variant, factory)
                return

class ObjectLen(FieldValidationMixin, schema.MinMaxLen, _ObjectBase):  # order matters
    """
    Allows specifying a length for arbitrary object fields (though the
    objects themselves must support the `len` function.
    """

    def __init__(self, sch, min_length=0, max_length=None, **kwargs):
        # match the calling sequence of Object, which uses a non-keyword
        # argument for schema.
        # But to work with the superclass, we have to pass it as a keyword arg.
        # it's weird.
        super(ObjectLen, self).__init__(schema=sch,
                                        min_length=min_length,
                                        max_length=max_length,
                                        **kwargs)

class Int(FieldValidationMixin, schema.Int):

    def fromUnicode(self, value):
        # Allow empty strings
        result = super(Int, self).fromUnicode(value) if value else None
        return result

class Float(FieldValidationMixin, schema.Float):

    def fromUnicode(self, value):
        result = super(Float, self).fromUnicode(value) if value else None
        return result

class Number(FieldValidationMixin, schema.Float):
    """
    A field that parses like a float from a string, but accepts any number.

    .. deprecated:: 1.4.0
       Use :class:`zope.schema.Number` if you really want arbitrary
       numbers (including fractions and decimals). You probably want
       :class:`zope.schema.Real` instead, though.
    """
    _type = numbers.Number

@__with_set()
class ValidChoice(FieldValidationMixin, schema.Choice):
    pass

@__with_set()
class ValidBytesLine(FieldValidationMixin, schema.BytesLine):
    pass

@__with_set()
class ValidBytes(FieldValidationMixin, schema.Bytes):
    pass

@__with_set(BeforeTextAssignedEvent)
class ValidText(FieldValidationMixin, schema.Text):
    """
    A text line that produces slightly better error messages. They will all
    have the 'field' property.

    We also fire :class:`IBeforeTextAssignedEvent`, which the normal
    mechanism does not.
    """

@__with_set(BeforeTextLineAssignedEvent)
class ValidTextLine(FieldValidationMixin, schema.TextLine):
    """
    A text line that produces slightly better error messages. They will all
    have the 'field' property.

    We also fire :class:`IBeforeTextLineAssignedEvent`, which the normal
    mechanism does not.
    """

@interface.implementer(sch_interfaces.IFromBytes)
class DecodingValidTextLine(ValidTextLine):
    """
    A text type that will attempt to decode non-unicode
    data as UTF-8.

    This primarily exists for legacy support (tests and persisted data).
    """

    def validate(self, value):
        if isinstance(value, bytes):
            value = value.decode('utf-8')  # let raise UnicodeDecodeError
        super(DecodingValidTextLine, self).validate(value)
        return value # tests

    # fromUnicode calls validate, so no need to duplicate

    def fromBytes(self, value):
        return self.fromUnicode(value.decode('utf-8'))

class ValidRegularExpression(ValidTextLine):

    def __init__(self, pattern, flags=(re.U|re.I|re.M), *args, **kwargs):
        super(ValidRegularExpression, self).__init__(*args, **kwargs)
        self.flags = flags
        self.pattern = pattern
        self.prog = re.compile(pattern, flags)

    def constraint(self, value): # pylint:disable=method-hidden
        # If they pass a 'constraint' kwarg, it will override this.
        return self.prog.match(value) is not None

ValidRegEx = ValidRegularExpression

class ValidURI(FieldValidationMixin, schema.URI):

    def _fixup_validation_error_args(self, e, value):
        if isinstance(e, sch_interfaces.InvalidURI):
            # This class differs by using the value as the argument, not
            # a message
            e.__doc__ = e.__doc__.replace('URI', 'URL')
            e.args = (value, e.__doc__, self.__fixup_name__)
            e.message = e.i18n_message = e.__doc__
        else: # pragma: no cover
            super(ValidURI, self)._fixup_validation_error_args(e, value)

class HTTPURL(ValidURI):
    """
    A URI field that ensures and requires its value to be an absolute
    HTTP/S URL.
    """

    def fromUnicode(self, value):
        # This can wind up producing something invalid if an
        # absolute URI was already given for mailto: for whatever.
        # None of the regexs (zopes or grubers) flag that as invalid.
        # so we try to
        orig_value = value
        if value:
            lower = value.lower()
            if not lower.startswith(u'http://') and not lower.startswith(u'https://'):
                # assume http
                value = u'http://' + value
        result = super(HTTPURL, self).fromUnicode(value)
        if result.count(u':') != 1:
            self._reraise_validation_error(
                sch_interfaces.InvalidURI(orig_value).with_field_and_value(self, orig_value),
                orig_value,
                _raise=True)

        return result

class _ValueTypeAddingDocMixin(object):
    """
    A mixin for fields that wrap a value type field (e.g., Object)
    to copy the nested documentation to the parent so it is visible
    in :mod:`repoze.sphinx.autointerface`.
    """

    def getExtraDocLines(self):
        lines = super(_ValueTypeAddingDocMixin, self).getExtraDocLines()
        accept_types = getattr(self, 'accept_types', None)
        if accept_types:
            # Private helper. If it goes away or changes, making sphinx docs will
            # fail, but we shouldn't have any runtime problems.
            from zope.schema._bootstrapfields import _DocStringHelpers
            lines.append(_DocStringHelpers.make_class_field('Accepted Types', accept_types))
        return lines


@__with_set(BeforeSequenceAssignedEvent)
class IndexedIterable(_ValueTypeAddingDocMixin, FieldValidationMixin, schema.Sequence):
    """
    An arbitrary (indexable) iterable, not necessarily a list or tuple;
    either of those would be acceptable at any time (however, so would a string,
    so be careful. Try ListOrTuple if that's a problem).

    The values may be homogeneous by setting the value_type.

    .. versionchanged:: 1.4.0
       Subclass :class:`zope.schema.Sequence` instead of :class:`zope.schema.List`,
       which adds checking that the value is indeed a sequence.
    """


@interface.implementer(IListOrTuple)
class ListOrTuple(IndexedIterable):
    "Restrict sequence values specifically to list and tuple."
    _type = (list, tuple)



@interface.implementer(IFromObject)
class _SequenceFromObjectMixin(object):
    accept_types = None
    _default_type = list

    def __init__(self, *args, **kwargs):
        super(_SequenceFromObjectMixin, self).__init__(*args, **kwargs)
        self._validate_contained_field(self.value_type)

    @classmethod
    def _validate_contained_field(cls, field):
        field = _fixup_Object_field(field)
        # We allow None for a value type
        if field is not None and not any(
                (iface.providedBy(field)
                 for iface in (IFromObject, IFromUnicode, IFromBytes))
        ):
            raise sch_interfaces.SchemaNotProvided(IFromObject, field)

    def _converter_for(self, field):
        return _FieldConverter(field)

    def _do_fromObject(self, context):
        converter = self._converter_for(self.value_type)
        result = [converter(x) for x in context]
        return result

    def _do_convert_result(self, result):
        if (isinstance(self._type, type)
                and self._type is not self._default_type):  # single type is a factory
            result = self._type(result)
        return result

    def fromObject(self, context):
        if context == self.missing_value:
            if self.required:
                raise sch_interfaces.RequiredMissing(self.__name__).with_field_and_value(self, context)
            # Nothing useful to do here, we're not required, don't
            # even try to convert
            return context

        check_type = self.accept_types or self._type
        if check_type is not None and not isinstance(context, check_type):
            raise sch_interfaces.WrongType(context, check_type).with_field_and_value(self, context)

        result = self._do_fromObject(context)
        return self._do_convert_result(result)


class _SequenceFromObject(_SequenceFromObjectMixin):

    def __init__(self, field):
        # We deliberately do not call the super init. The field's value_type
        # must already be valid.
        # pylint:disable=super-init-not-called
        self.__field = field

    def __getattr__(self, name):
        return getattr(self.__field, name)


class _MapFromObjectMixin(_SequenceFromObjectMixin):

    def __init__(self, *args, **kwargs):
        super(_MapFromObjectMixin, self).__init__(*args, **kwargs)
        self._validate_contained_field(self.key_type)

    def _do_convert_result(self, result):
        assert isinstance(result, dict)
        return result

    def _do_fromObject(self, context):
        key_converter = self._converter_for(self.key_type)
        value_converter = self._converter_for(self.value_type)
        return {key_converter(k): value_converter(v) for k, v in _iteritems(context)}


class _MapFromObject(_MapFromObjectMixin):

    def __init__(self, field):
        # We deliberately do not call the super init. The field's value_type
        # and key_type must already be valid.
        # pylint:disable=super-init-not-called
        self.__field = field

    def __getattr__(self, name):
        return getattr(self.__field, name)


class ListOrTupleFromObject(_SequenceFromObjectMixin, ListOrTuple):
    """
    The ``value_type`` MUST be a :class:`Variant`, or more generally,
    something supporting :class:`IFromObject`, :class:`IFromUnicode`
    or :class:`IFromBytes`.

    .. versionchanged:: 1.9.0

        Implementations of :class:`zope.schema.interfaces.ICollection`
        (like :class:`zope.schema.List`) and
        :class:`zope.schema.interfaces.IMapping` (like
        :class:`zope.schema.Dict`) will automatically be given a ``fromObject`` method
        when used as the *value_type* of this object *if* their *value_type* is an
        :class:`zope.schema.interfaces.IObject` (recursively).
    """


class TupleFromObject(_ValueTypeAddingDocMixin,
                      _SequenceFromObjectMixin,
                      FieldValidationMixin,
                      schema.Tuple):
    """
    The ``value_type`` MUST be a :class:`Variant`, or more generally,
    something supporting :class:`IFromObject`, :class:`IFromUnicode`
    or :class:`IFromBytes`.

    When setting through this object, we will automatically convert
    lists and only lists to tuples (for convenience coming in through
    JSON).

    .. versionchanged:: 1.9.0

        Implementations of :class:`zope.schema.interfaces.ICollection`
        (like :class:`zope.schema.List`) and
        :class:`zope.schema.interfaces.IMapping` (like
        :class:`zope.schema.Dict`) will automatically be given a ``fromObject`` method
        when used as the *value_type* of this object *if* their *value_type* is an
        :class:`zope.schema.interfaces.IObject` (recursively).
    """
    accept_types = (list, tuple)

    def set(self, context, value):
        if isinstance(value, list):
            value = tuple(value)

        _do_set(self, context, value, TupleFromObject, BeforeSequenceAssignedEvent)

    def validate(self, value):
        if isinstance(value, list):
            value = tuple(value)
        super(TupleFromObject, self).validate(value)


@__with_set(BeforeDictAssignedEvent)
class DictFromObject(_ValueTypeAddingDocMixin,
                     _MapFromObjectMixin,
                     FieldValidationMixin,
                     schema.Mapping):
    """
    The `key_type` and `value_type` must be supporting
    :class:`IFromObject` or :class:`.IFromUnicode` or :class:`IFromBytes`

    .. versionchanged:: 1.4.0
       Subclass :class:`zope.schema.Mapping` instead of :class:`zope.schema.Dict`,
       allowing for any mapping (such as BTrees), not just dicts. However, the validated
       value is still a dict.

    .. versionchanged:: 1.9.0

        Implementations of :class:`zope.schema.interfaces.ICollection`
        (like :class:`zope.schema.List`) and
        :class:`zope.schema.interfaces.IMapping` (like
        :class:`zope.schema.Dict`) will automatically be given a ``fromObject`` method
        when used as the *value_type* of this object *if* their *value_type* is an
        :class:`zope.schema.interfaces.IObject` (recursively).
    """


@__with_set(BeforeSetAssignedEvent)
class ValidSet(_ValueTypeAddingDocMixin,
               _SequenceFromObjectMixin,
               FieldValidationMixin,
               schema.Set):
    """
    A set that is validated.

    .. versionchanged:: 1.5.0
       Implement :class:`IFromObject`.

    .. versionchanged:: 1.9.0

        Implementations of :class:`zope.schema.interfaces.ICollection`
        (like :class:`zope.schema.List`) and
        :class:`zope.schema.interfaces.IMapping` (like
        :class:`zope.schema.Dict`) will automatically be given a ``fromObject`` method
        when used as the *value_type* of this object *if* their *value_type* is an
        :class:`zope.schema.interfaces.IObject` (recursively).
    """


class UniqueIterable(ValidSet):
    """
    An arbitrary iterable, not necessarily an actual :class:`set` object,
    but one whose contents are unique. Use this when you can
    return a :class:`set`, :class:`frozenset` or empty tuple. These should be
    sequences that suport the ``in`` operator.

    .. versionchanged:: 1.5.0
       Implement :class:`IFromObject`.

    .. versionchanged:: 1.9.0

        Implementations of :class:`zope.schema.interfaces.ICollection`
        (like :class:`zope.schema.List`) and
        :class:`zope.schema.interfaces.IMapping` (like
        :class:`zope.schema.Dict`) will automatically be given a ``fromObject`` method
        when used as the *value_type* of this object *if* their *value_type* is an
        :class:`zope.schema.interfaces.IObject` (recursively).
    """
    _type = None  # Override to not force a set

    def __init__(self, *args, **kwargs):
        # If they do not specify a min_length in the arguments,
        # then change it to None. This way we are compatible with
        # a generator value. Superclass specifies both a class value
        # and a default argument
        no_min_length = False
        if 'min_length' not in kwargs:
            no_min_length = True

        super(UniqueIterable, self).__init__(*args, **kwargs)
        if no_min_length:
            self.min_length = None

    def _do_convert_result(self, result):
        if not isinstance(result, (set, frozenset)):
            result = set(result)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for field.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import unittest
import warnings


from zope.component import eventtesting

from zope.interface.common import interfaces as cmn_interfaces
from zope.schema import Dict
from zope.schema.interfaces import ValidationError
from zope.schema.interfaces import InvalidURI
from zope.schema.interfaces import SchemaNotProvided
from zope.schema.interfaces import TooLong
from zope.schema.interfaces import TooShort
from zope.schema.interfaces import WrongType


from nti.schema.field import HTTPURL
from nti.schema.field import DecodingValidTextLine
from nti.schema.field import FieldValidationMixin
from nti.schema.field import Float
from nti.schema.field import IndexedIterable
from nti.schema.field import Int
from nti.schema.field import ListOrTuple
from nti.schema.field import ListOrTupleFromObject
from nti.schema.field import Number
from nti.schema.field import Object
from nti.schema.field import ObjectLen
from nti.schema.field import TupleFromObject
from nti.schema.field import UniqueIterable
from nti.schema.field import ValidDatetime
from nti.schema.field import ValidRegularExpression
from nti.schema.field import Variant
from nti.schema.field import ValidTextLine as TextLine
from nti.schema.interfaces import IBeforeDictAssignedEvent
from nti.schema.interfaces import IBeforeSequenceAssignedEvent
from nti.schema.interfaces import IVariant
from nti.schema.interfaces import IFromObject
from nti.schema.interfaces import VariantValidationError
# Import from the BWC location

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from nti.schema.testing import validated_by # pylint:disable=no-name-in-module
    from nti.schema.testing import verifiably_provides # pylint:disable=no-name-in-module
    from nti.schema.testing import not_validated_by # pylint:disable=no-name-in-module


from . import IUnicode
from . import SchemaLayer

from hamcrest import assert_that
from hamcrest import calling
from hamcrest import contains
from hamcrest import contains_string
from hamcrest import has_length
from hamcrest import has_property
from hamcrest import is_
from hamcrest import is_not
from hamcrest import none
from hamcrest import raises

does_not = is_not

__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904,blacklisted-name


class TestObjectLen(unittest.TestCase):


    def test_objectlen(self):
        # If we have the inheritance messed up, we will have problems
        # creating, or we will have problems validating one part or the other.

        olen = ObjectLen(IUnicode, max_length=5)  # default val for min_length

        olen.validate(u'a')
        olen.validate(u'')

        assert_that(calling(olen.validate).with_args(object()),
                    raises(SchemaNotProvided))

        assert_that(calling(olen.validate).with_args(u'abcdef'),
                    raises(TooLong))

    def test_objectlen_short(self):
        olen = ObjectLen(IUnicode, min_length=5)

        assert_that(calling(olen.validate).with_args(u'abc'),
                    raises(TooShort))

class TestHTTPUrl(unittest.TestCase):

    def test_http_url(self):

        http = HTTPURL(__name__='foo')

        assert_that(http.fromUnicode('www.google.com'),
                    is_('http://www.google.com'))

        assert_that(http.fromUnicode('https://www.yahoo.com'),
                    is_('https://www.yahoo.com'))

        with self.assertRaises(InvalidURI) as exc:
            http.fromUnicode('mailto:jason@nextthought.com')

        exception = exc.exception
        assert_that(exception, has_property('field', http))
        assert_that(exception, has_property('value', 'mailto:jason@nextthought.com'))
        assert_that(exception, has_property('message', 'The specified URL is not valid.'))

class TestVariant(unittest.TestCase):

    def test_variant(self):

        syntax_or_lookup = Variant((Object(cmn_interfaces.ISyntaxError),
                                    Object(cmn_interfaces.ILookupError),
                                    Object(IUnicode)))

        assert_that(syntax_or_lookup, verifiably_provides(IVariant))

        # validates
        assert_that(SyntaxError(), validated_by(syntax_or_lookup))
        assert_that(LookupError(), validated_by(syntax_or_lookup))

        # doesn't validate
        assert_that(b'foo', not_validated_by(syntax_or_lookup))

        assert_that(syntax_or_lookup.fromObject(u'foo'), is_(u'foo'))

        with self.assertRaises(VariantValidationError) as exc:
            syntax_or_lookup.fromObject(object())

        assert_that(exc.exception,
                    has_property('errors',
                                 has_length(len(syntax_or_lookup.fields))))

    def test_getDoc(self):
        syntax_or_lookup = Variant((Object(cmn_interfaces.ISyntaxError),
                                    Object(cmn_interfaces.ILookupError),
                                    Object(IUnicode)))

        doc = syntax_or_lookup.getDoc()
        assert_that(doc, contains_string('.. rubric:: Possible Values'))
        assert_that(doc, contains_string('.. rubric:: Option'))

    def test_complex_variant(self):

        dict_field = Dict(key_type=TextLine(), value_type=TextLine())
        string_field = Object(IUnicode)
        list_of_numbers_field = ListOrTuple(value_type=Number())

        variant = Variant((dict_field, string_field, list_of_numbers_field))
        variant.getDoc()  # cover
        # It takes all these things
        for d in {u'k': u'v'}, u'foo', [1, 2, 3]:
            assert_that(d, validated_by(variant))

        # It rejects these by raising a VariantValidationError
        # with the same number of errors as fields
        for d in {u'k': 1}, b'foo', [1, 2, u'b']:
            assert_that(d, not_validated_by(variant))
            with self.assertRaises(VariantValidationError) as exc:
                variant.fromObject(d)

            assert_that(exc.exception, has_property("errors",
                                                    has_length(len(variant.fields))))

            with self.assertRaises(VariantValidationError) as exc:
                variant.validate(d)

            assert_that(exc.exception, has_property("errors",
                                                    has_length(len(variant.fields))))

        # A name set now is reflected down the line
        variant.__name__ = 'baz'
        for f in variant.fields:
            assert_that(f, has_property('__name__', 'baz'))

        # and in clones
        clone = variant.bind(object())
        for f in clone.fields:
            assert_that(f, has_property('__name__', 'baz'))

        # which doesn't change the original
        clone.__name__ = 'biz'
        for f in clone.fields:
            assert_that(f, has_property('__name__', 'biz'))

        for f in variant.fields:
            assert_that(f, has_property('__name__', 'baz'))

        # new objects work too
        new = Variant(variant.fields, __name__='boo')
        for f in new.fields:
            assert_that(f, has_property('__name__', 'boo'))

    def test_variant_from_object(self):
        field = Variant((TupleFromObject(HTTPURL()),))

        res = field.fromObject(['http://example.com'])
        assert_that(res, is_(('http://example.com',)))

    def test_variant_auto_convert_sequence(self):
        from zope.schema import List
        from zope.schema import Field

        string_field = Object(IUnicode)
        list_field = List(string_field)
        assert_that(list_field, does_not(verifiably_provides(IFromObject)))

        # But once we add it to a variant, it does

        variant = Variant((list_field,))

        assert_that(list_field, verifiably_provides(IFromObject))

        bound_variant = variant.bind(self)

        assert_that(bound_variant.fields[0], verifiably_provides(IFromObject))

        l = bound_variant.fromObject([u'abc'])
        assert_that(l, is_([u'abc']))

        # Doing the same with a list of a field we don't know about
        # does nothing.
        field = Field()
        list_field = List(field)
        assert_that(list_field, does_not(verifiably_provides(IFromObject)))
        variant = Variant((list_field,))
        assert_that(list_field, does_not(verifiably_provides(IFromObject)))

    def test_variant_auto_convert_mapping(self):
        from zope.schema import Dict as ZDict # pylint:disable=reimported
        from zope.schema import Field

        string_field = Object(IUnicode)
        map_field = ZDict(string_field, string_field)
        assert_that(map_field, does_not(verifiably_provides(IFromObject)))

        # But once we add it to a variant, it does

        variant = Variant((map_field,))

        assert_that(map_field, verifiably_provides(IFromObject))

        bound_variant = variant.bind(self)

        assert_that(bound_variant.fields[0], verifiably_provides(IFromObject))

        l = bound_variant.fromObject({u'abc': u'def'})
        assert_that(l, is_({u'abc': u'def'}))

        # Doing the same with a list of a field we don't know about
        # does nothing.
        map_field = ZDict(Field(), Field())
        assert_that(map_field, does_not(verifiably_provides(IFromObject)))
        variant = Variant((map_field,))
        assert_that(map_field, does_not(verifiably_provides(IFromObject)))

    def test_converts_but_not_valid(self):
        # If the schema accepts the input, but the validation refuses,
        # keep going.
        class WeirdField(Object):
            schema = IUnicode

            def validate(self, value):
                raise SchemaNotProvided().with_field_and_value(self, value)
        weird_field = WeirdField(IUnicode)
        accept_field = Number()

        field = Variant((weird_field, accept_field),
                        variant_raise_when_schema_provided=True)
        assert_that(field.fromObject("1.0"),
                    is_(1.0))

        assert_that(calling(field.validate).with_args(u'1.0'),
                    raises(SchemaNotProvided))

    def test_invalid_construct(self):
        assert_that(calling(Variant).with_args(()),
                    raises(SchemaNotProvided))

    def test_AttributeError_transformed(self):
        # Something that blows up badly on None input
        # to fromUnicode never gets there.
        from zope.schema.interfaces import RequiredMissing
        http_field = HTTPURL()
        with self.assertRaises(AttributeError):
            http_field.fromUnicode(None)

        # Gets wrapped up into a validation error by a Variant
        # (anything that does a conversion, actually).
        # We have to use a non-default (and thus non None)
        # missing value to take this code path.
        field = Variant((http_field,),
                        missing_value=self)

        with self.assertRaises(VariantValidationError) as exc:
            field.fromObject(None)

        ex = exc.exception
        assert_that(ex.errors, has_length(1))
        assert_that(ex.errors[0], is_(RequiredMissing))

        for val in repr(ex), str(ex):
            assert_that(val, contains_string("RequiredMissing"))
            assert_that(val, contains_string("Value:"))
            assert_that(val, contains_string("Field:"))

    def test_numbers_pass_unchanged(self):
        from zope.schema import Number as ZNumber

        number = ZNumber()
        variant = Variant([number])

        assert_that(variant.fromObject(10), is_(10))

    def test_None_not_passed_to_fromUnicode(self):
        # But something that can accept None
        # does get it still
        from zope.schema import Field
        class MyField(Field):

            last_value = None

            def fromUnicode(self, value):
                raise AssertionError("Should not be called")

            def validate(self, value):
                self.last_value = 42

        field = MyField()
        variant = Variant([field], required=False)
        x = variant.fromObject(None)
        assert_that(x, is_(none()))
        assert_that(field.last_value, is_(42))

    def test_missing_value_not_required(self):
        from zope.schema.interfaces import RequiredMissing
        text = TextLine()

        variant = Variant((text,), required=False)
        # None is the default missing value, and it's allowed
        assert_that(variant.fromObject(None),
                    is_(none()))

        # If we are required, though, we cannot use the missing value
        variant = Variant((text,), required=True)
        with self.assertRaises(RequiredMissing):
            variant.fromObject(None)

        # We can use arbitrary other missing values
        variant = Variant((text,), required=False, missing_value=self)
        assert_that(variant.fromObject(self),
                    is_(self))

        # And None now raises.
        with self.assertRaises(VariantValidationError):
            variant.fromObject(None)

class TestConfiguredVariant(unittest.TestCase):

    layer = SchemaLayer

    def test_nested_variants(self):
        # Use case: Chat messages are either a Dict, or a N
        # ote-like body, which itself is a list of variants

        dict_field = Dict(key_type=TextLine(), value_type=TextLine())

        string_field = Object(IUnicode)
        number_field = Number()
        list_of_strings_or_numbers = ListOrTuple(value_type=Variant((string_field, number_field)))

        assert_that([1, u'2'], validated_by(list_of_strings_or_numbers))
        assert_that({u'k': u'v'}, validated_by(dict_field))

        dict_or_list = Variant((dict_field, list_of_strings_or_numbers))

        assert_that([1, u'2'], validated_by(dict_or_list))
        assert_that({u'k': u'v'}, validated_by(dict_or_list))

        class X(object):
            pass

        x = X()
        dict_or_list.set(x, [1, u'2'])

        events = eventtesting.getEvents(IBeforeSequenceAssignedEvent)
        assert_that(events, has_length(1))
        assert_that(events, contains(has_property('object', [1, '2'])))

        eventtesting.clearEvents()

        dict_or_list.set(x, {u'k': u'v'})
        events = eventtesting.getEvents(IBeforeDictAssignedEvent)
        assert_that(events, has_length(1))
        assert_that(events, contains(has_property('object', {'k': 'v'})))

class TestValidSet(unittest.TestCase):

    def _getTargetClass(self):
        from nti.schema.field import ValidSet
        return ValidSet

    def _makeOne(self):
        return self._getTargetClass()()

    default_min_length = 0

    def test_min_length(self):
        field = self._getTargetClass()(__name__='foo')
        assert_that(field, has_property('min_length',
                                        self.default_min_length))

        class Thing(object):
            foo = None
        thing = Thing()
        field.set(thing, ())

        assert_that(thing, has_property('foo', ()))

    def test_value_type(self):
        field = self._getTargetClass()(value_type=HTTPURL())

        o = field.fromObject({'https://example.com'})
        assert_that(o, is_({'https://example.com'}))

    def test_interfaces(self):
        assert_that(self._makeOne(),
                    verifiably_provides(IFromObject))


class TestUniqueIterable(TestValidSet):

    def _getTargetClass(self):
        return UniqueIterable

    default_min_length = none()

class SequenceFromObjectMixinMixin(object):

    def _getTargetClass(self):
        raise NotImplementedError

    def _makeOne(self, *args, **kwargs):
        return self._getTargetClass()(*args, **kwargs)

    def test_accepts_missing_value(self):
        from zope.schema.interfaces import RequiredMissing
        field = self._makeOne(required=False, missing_value=None)
        assert_that(field.fromObject(None), is_(none()))

        # but if we're required, we get a different exception
        field = self._makeOne(required=True, missing_value=None)
        with self.assertRaises(RequiredMissing):
            field.fromObject(None)

        # We can change the type
        field = self._makeOne(required=False, missing_value=self)
        assert_that(field.fromObject(self), is_(self))


class TestTupleFromObject(SequenceFromObjectMixinMixin,
                          unittest.TestCase):

    def _getTargetClass(self):
        return TupleFromObject

    def test_set(self):
        field = TupleFromObject(__name__='foo')

        class Thing(object):
            foo = None

        thing = Thing()
        field.validate([1, 2])
        field.set(thing, [1, 2])
        assert_that(thing, has_property('foo', (1, 2)))

        # But arbitrary iterables not validated...
        assert_that(calling(field.validate).with_args('abc'),
                    raises(WrongType))

        # Although they can be set...
        field.set(thing, 'abc')

    def test_wrong_type_from_object(self):
        field = TupleFromObject()
        assert_that(calling(field.fromObject).with_args('abc'),
                    raises(WrongType))

    def test_valid_type_from_object_unicode(self):
        field = TupleFromObject(HTTPURL())
        res = field.fromObject(['http://example.com'])
        assert_that(res, is_(('http://example.com',)))

    def test_valid_type_from_object_object(self):
        # Nested layers of fromObject and fromUnicode
        field = TupleFromObject(Variant((HTTPURL(),)))
        res = field.fromObject(['http://example.com'])
        assert_that(res, is_(('http://example.com',)))



class TestListOrTupleFromObject(SequenceFromObjectMixinMixin,
                                unittest.TestCase):

    def _getTargetClass(self):
        return ListOrTupleFromObject

    def test_construct_with_object(self):
        field = ListOrTupleFromObject(Object(IUnicode))

        assert_that(field.value_type, verifiably_provides(IFromObject))

        # And it tries to adapt individual fields
        with self.assertRaises(TypeError) as exc:
            field.fromObject((b'abc',))

        exception = exc.exception
        self.assertEqual(exception.args[0], 'Could not adapt')

        class Conforms(object):

            def __conform__(self, iface):
                assert_that(iface, is_(IUnicode))
                return u'def'

        o = field.fromObject((Conforms(),))
        assert_that(o, is_([u'def']))

        # This works even when bound

        field = field.bind(self)
        assert_that(field.value_type, verifiably_provides(IFromObject))
        o = field.fromObject((Conforms(),))
        assert_that(o, is_([u'def']))

    def test_construct_with_bad_field(self):

        from zope.schema import Field

        with self.assertRaises(SchemaNotProvided):
            ListOrTupleFromObject(Field())


class TestIndexedIterable(unittest.TestCase):

    def test_accepts_str(self):
        field = IndexedIterable(__name__='foo')
        class Thing(object):
            foo = None

        thing = Thing()
        field.set(thing, 'abc')
        assert_that(thing, has_property('foo', 'abc'))

class TestDecodingValidTextLine(unittest.TestCase):

    def test_decode(self):
        field = DecodingValidTextLine()
        res = field.validate(b'abc')
        assert_that(res, is_(u'abc'))

    def test_fromBytes(self):
        from zope.schema.interfaces import IFromBytes
        field = DecodingValidTextLine()
        assert_that(field, verifiably_provides(IFromBytes))

        val = field.fromBytes(b'abc')
        assert_that(val, is_(u'abc'))

class TestNumber(unittest.TestCase):

    def test_allow_empty(self):
        assert_that(Float().fromUnicode(''), is_(none()))
        assert_that(Int().fromUnicode(''), is_(none()))

class TestDatetime(unittest.TestCase):

    def test_validate_wrong_type(self):
        field = ValidDatetime()
        assert_that(calling(field.validate).with_args(''),
                    raises(SchemaNotProvided))

class TestFieldValidationMixin(unittest.TestCase):

    def test_one_arg(self):
        field = FieldValidationMixin()
        field.__name__ = 'foo'

        ex = SchemaNotProvided('msg').with_field_and_value(field, 'value')
        # zope.schema 4.7 automatically fills in args
        assert_that(ex.args, is_(('msg', None)))
        assert_that(ex.schema, is_('msg'))
        ex.args = ('msg',)
        try:
            field._reraise_validation_error(ex, 'value', _raise=True)
        except SchemaNotProvided:
            assert_that(ex.args, is_(('value', 'msg', 'foo')))

    def test_no_arg(self):
        field = FieldValidationMixin()
        field.__name__ = 'foo'

        ex = SchemaNotProvided().with_field_and_value(field, 'value')
        # zope.schema 4.7 automatically fills in args
        assert_that(ex.args, is_((None, None)))
        ex.args = ()
        try:
            field._reraise_validation_error(ex, 'value', _raise=True)
        except SchemaNotProvided:
            assert_that(ex.args, is_(('value', '', 'foo')))

    def test_oob(self):
        from zope.schema import Integral
        from zope.schema.interfaces import OutOfBounds
        class Field(FieldValidationMixin, Integral):
            pass

        field = Field(min=1)
        with self.assertRaises(OutOfBounds):
            field.validate(0)

    def test_random_validation_error(self):
        class Field(object):
            __name__ = ''
            def _validate(self, v):
                raise ValidationError().with_field_and_value(self, v)

        class Field2(FieldValidationMixin, Field):
            pass

        field = Field2()
        with self.assertRaises(ValidationError) as exc:
            field._validate(42)

        ex = exc.exception
        assert_that(ex.args, is_((42, '', '')))


class TestRegex(unittest.TestCase):

    def test_regex(self):
        field = ValidRegularExpression('[bankai|shikai]', flags=0)
        assert_that(field.constraint("bankai"), is_(True))
        assert_that(field.constraint("shikai"), is_(True))
        assert_that(field.constraint("Shikai"), is_(False))
        assert_that(field.constraint("foo"), is_(False))
        field = ValidRegularExpression('[bankai|shikai]')
        assert_that(field.constraint("Shikai"), is_(True))
        assert_that(field.constraint("banKAI"), is_(True))


class TestValueTypeAddingDocMixin(unittest.TestCase):

    def _getTargetClass(self):
        from nti.schema.field import _ValueTypeAddingDocMixin
        return _ValueTypeAddingDocMixin

    def test_getDoc(self):

        from zope.schema import Field

        class MyField(self._getTargetClass(), Field):
            _type = object
            accept_types = (list, tuple)

        doc = MyField().getDoc()

        assert_that(doc, contains_string(':Allowed Type: :class:`object`'))
        assert_that(doc, contains_string(':Accepted Types: :class:`list`, :class:`tuple`'))


class TestDictFromObject(SequenceFromObjectMixinMixin,
                         unittest.TestCase):

    def _getTargetClass(self):
        from nti.schema.field import DictFromObject
        return DictFromObject

    def test_interfaces(self):
        assert_that(self._makeOne(),
                    verifiably_provides(IFromObject))

    def test_accepts_mapping(self):
        from six.moves import UserDict
        from nti.schema.field import abcs

        # On Python 2, UserDict is not registered as a Mapping.
        if not issubclass(UserDict, abcs.Mapping): # pragma: no cover
            abcs.Mapping.register(UserDict)

        field = self._makeOne(key_type=Int(), value_type=Float())

        value = UserDict({'1': '42'})
        assert_that(value, is_(abcs.Mapping))
        assert_that(value, is_not(dict))

        result = field.fromObject(value)
        assert_that(result, is_(dict))
        assert_that(result, is_({1: 42.0}))

    def test_getDoc(self):
        field = self._makeOne(key_type=Int(), value_type=Float())
        doc = field.getDoc()

        assert_that(doc, contains_string('.. rubric:: Value Type'))
        assert_that(doc, contains_string('.. rubric:: Key Type'))


class Test_FieldConverter(unittest.TestCase):

    def _getTargetClass(self):
        from nti.schema.field import _FieldConverter
        return _FieldConverter

    def _makeOne(self, field=None):
        return self._getTargetClass()(field)

    def test_fromBytes(self):

        class Field(object):
            def fromBytes(self, value):
                return b'from bytes'

            def validate(self, value):
                raise WrongType

        field = Field()
        converter = self._makeOne(field)
        assert_that(converter(b''), is_(b'from bytes'))

        with self.assertRaises(WrongType) as exc:
            converter(1)

        exception = exc.exception
        assert_that(exception, has_property('field', field))
        assert_that(exception, has_property('value', 1))

    def test_validate_called_last(self):
        # If we had nothing else to do, we
        # call validate and if it doesn't raise, we return
        class Field(object):
            def validate(self, value):
                if value == 42:
                    raise ValidationError

        field = Field()
        # A bad value raises, and we fill in the field and value if
        # needed.
        converter = self._makeOne(field)
        with self.assertRaises(ValidationError) as exc:
            converter(42)

        assert_that(exc.exception.field, is_(field))
        assert_that(exc.exception.value, is_(42))

        # Otherwise we return the value we were passed
        assert_that(converter(b''), is_(b''))
        assert_that(converter(u''), is_(u''))

    def test_fromObject(self):
        # from object is only called for non-strings if
        # fromUnicode and fromBytes are implemented.
        class Field(object):
            def fromObject(self, value):
                return b'from object'

            def fromUnicode(self, value):
                return u'from unicode'

            def fromBytes(self, value):
                return b'from bytes'

        converter = self._makeOne(Field())
        assert_that(converter(b''), is_(b'from bytes'))
        assert_that(converter(u''), is_(u'from unicode'))
        assert_that(converter(1), is_(b'from object'))



#!/usr/bin/env python

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import unittest

from zope.interface import Interface

from nti.schema.jsonschema import JsonSchemafier

from . import SchemaLayer

from hamcrest import assert_that
from hamcrest import has_entry
from hamcrest import has_item
from hamcrest import has_property
from hamcrest import is_
from hamcrest import is_in
from hamcrest import is_not
from hamcrest import not_none

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904,inherit-non-class


does_not = is_not


class TestConfiguredVocabulary(unittest.TestCase):

    layer = SchemaLayer

    def test_country_vocabulary(self):
        from zope.schema import Choice

        class IA(Interface):
            choice = Choice(title=u"Choice",
                            vocabulary=u"Countries")

        o = object()

        choice = IA['choice'].bind(o)
        assert_that(choice.vocabulary, is_(not_none()))
        assert_that('us', is_in(choice.vocabulary))
        term = choice.vocabulary.getTermByToken('us')
        assert_that(term, has_property('value', "United States"))
        ext = term.toExternalObject()
        assert_that(ext, has_entry('flag', u'/++resource++country-flags/us.gif'))
        assert_that(ext, has_entry('title', 'United States'))

        schema = JsonSchemafier(IA).make_schema()
        assert_that(schema, has_entry('choice', has_entry('choices', has_item(ext))))

def test_suite():
    import doctest
    suite = unittest.defaultTestLoader.loadTestsFromName(__name__)
    suite.addTest(
        doctest.DocTestSuite("nti.schema.vocabulary")
    )

    return suite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helpers for writing code that implements schemas.

This module contains code based on code originally from dm.zope.schema.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope import interface
from zope.deferredimport import deprecatedFrom
from zope.interface import Interface
from zope.interface import providedBy
from zope.schema import getFieldsInOrder

from .interfaces import ISchemaConfigured

__docformat__ = "restructuredtext en"

def _interface_from_spec(spec):
    """
    build an interface from interface spec *spec*.

    This should probably go into its own package (maybe ``dm.zope.interface``).

    Note: the interfaces constructed in this way may not be picklable
    (not tested). If they are indeed not picklable, they should not be stored.
    """
    # try to guess whether "spec" is an interface or specification
    #  The implementors have broken "issubclass", we therefore need
    #  more indirect (and in principle less reliable) methods
    # if issubclass(spec, Interface):
    # For example, the implementedBy object tests true, but can't be passed to getFieldsInOrder

    if hasattr(spec, 'names'):
        return spec # pragma: no cover
    return type(Interface)('FromSpec', tuple(spec), {})


def schemaitems(spec):
    """The schema part of interface specification *spec* as a list of id, field pairs."""
    iface = _interface_from_spec(spec)
    # may want to filter duplicates out or raise an exception on duplicates
    seen = set()
    items = []
    for (name, field) in getFieldsInOrder(iface):
        if name in seen:
            continue # pragma: no cover
        seen.add(name)
        items.append((name, field))
    return items


def schemadict(spec):
    """The schema part of interface specification *spec* as a ``dict``."""
    return dict(schemaitems(spec))


_marker = object()


@interface.implementer(ISchemaConfigured)
class SchemaConfigured(object):
    """Mixin class to provide configuration by the provided schema components."""

    def __init__(self, **kw):
        schema = schemadict(self.sc_schema_spec())
        for k, v in kw.items():
            # might want to control this check
            if k not in schema:
                raise TypeError('non schema keyword argument: %s' % k)
            setattr(self, k, v)
        # provide default values for schema fields not set
        for f, fields in schema.items():
            if getattr(self, f, _marker) is _marker:
                # The point of this is to avoid hiding exceptions (which the builtin
                # hasattr() does on Python 2)
                setattr(self, f, fields.default)

    # provide control over which interfaces define the data schema
    SC_SCHEMAS = None

    def sc_schema_spec(self):
        """the schema specification which determines the data schema.

        This is determined by `SC_SCHEMAS` and defaults to `providedBy(self)`.
        """
        spec = self.SC_SCHEMAS
        return spec or providedBy(self)

class PermissiveSchemaConfigured(SchemaConfigured):
    """
    A mixin subclass of :class:`SchemaConfigured` that allows
    for extra keywords (those not defined in the schema) to silently be ignored.
    This is an aid to evolution of code and can be helpful in testing.

    To allow for one-by-one conversions and updates, this class defines an attribute
    ``SC_PERMISSIVE``, defaulting to True, that controls this behaviour.
    """

    SC_PERMISSIVE = True

    def __init__(self, **kwargs):
        if not self.SC_PERMISSIVE:
            super(PermissiveSchemaConfigured, self).__init__(**kwargs)
        else:
            _schema = schemadict(self.sc_schema_spec())
            kwargs = {k: kwargs[k] for k in kwargs if k in _schema}
            super(PermissiveSchemaConfigured, self).__init__(**kwargs)


deprecatedFrom("Moved to nti.schema.eqhash",
               "nti.schema.eqhash",
               'EqHash',
               '_superhash')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for interfaces.py

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import unittest

from zope.deprecation import Suppressor

with Suppressor():
    from ..interfaces import InvalidValue

from hamcrest import assert_that
from hamcrest import has_property
from hamcrest import none
from hamcrest import is_
from hamcrest import instance_of

__docformat__ = "restructuredtext en"

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904

class TestInvalidValue(unittest.TestCase):

    def test_construct(self):
        v = InvalidValue()
        assert_that(v, has_property('value', none()))
        assert_that(v, has_property('field', none()))

        v = InvalidValue(value=1, field=1)
        assert_that(v, has_property('value', 1))
        assert_that(v, has_property('field', 1))

        with self.assertRaises(TypeError):
            InvalidValue(value=1, field=2, other=3)

    def test_subclass(self):
        kind = type('subclass', (InvalidValue,), {})
        self.test_instance(kind)
        self.test_raise_except(kind)

    def test_instance(self, kind=InvalidValue):
        from zope.schema import interfaces as sch_interfaces
        v = kind()
        self.assertIsInstance(v, sch_interfaces.InvalidValue)
        self.assertIsInstance(v, InvalidValue)

    def test_raise_except(self, kind=InvalidValue):
        # It can be caught by its own name
        with self.assertRaises(kind):
            raise kind()
        # It can be caught by the zope.schema name
        from zope.schema import interfaces as sch_interfaces
        with self.assertRaises(sch_interfaces.InvalidValue):
            raise InvalidValue()

    def test_repr(self):
        assert_that(repr(InvalidValue('foo')),
                    is_("InvalidValue('foo')"))

    def test_alias(self):
        from zope.schema import interfaces as sch_interfaces
        self.assertIs(InvalidValue, sch_interfaces.InvalidValue)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Benchmark for eqhash.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..eqhash import EqHash

@EqHash('a', 'b')
class Thing(object):
    a = 'a'
    b = 'b'

@EqHash('a', 'b', superhash=True)
class Thing2(object):
    a = 'a'
    b = 'b'

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

@EqHash('a', 'b', include_type=True)
class NotThing(object):
    a = 'a'
    b = 'b'

@EqHash('c', include_super=True)
class ChildThing(Thing):
    c = 'c'

@EqHash('c', include_super=False)
class ChildThingNoSuper(Thing):
    c = 'c'

@EqHash(include_super=True)
class ChildThingNoNames(Thing):
    pass

@EqHash('a', 'b', 'c', 'd', 'e', 'f')
class ManyThing(object):
    a = 'a'
    b = 'b'
    c = 'c'
    d = 'd'
    e = 'e'
    f = 'f'

# pylint:disable=line-too-long

def bench_hash(): # pragma: no cover
    import timeit
    import statistics


    timer = timeit.Timer('hash(thing)',
                         'from nti.schema.tests.eqhash_benchmark import Thing as Thing; thing=Thing()')
    times = timer.repeat()
    print("Avg Base  hash", statistics.mean(times), "stddev", statistics.stdev(times))

    timer = timeit.Timer('hash(thing)',
                         'from nti.schema.tests.eqhash_benchmark import ChildThing as Thing; thing=Thing()')
    times = timer.repeat()
    print("Avg Child hash", statistics.mean(times), "stddev", statistics.stdev(times))

    timer = timeit.Timer('hash(thing)', 'from nti.schema.tests.eqhash_benchmark import Thing2 as Thing; thing=Thing()')
    times = timer.repeat()
    print("Avg Super  hash", statistics.mean(times), "stddev", statistics.stdev(times))

    timer = timeit.Timer('hash(thing)', 'from nti.schema.tests.eqhash_benchmark import Thing2 as Thing; thing=Thing(a={})')
#    import cProfile
#    import pstats
#    pr = cProfile.Profile()
#    pr.enable()
    times = timer.repeat()
#    pr.disable()
#    ps = pstats.Stats(pr).sort_stats('cumulative')
#    ps.print_stats(.4)

    print("Avg Super2  hash", statistics.mean(times), "stddev", statistics.stdev(times))


def bench_eq(): # pragma: no cover
    import timeit
    import statistics


    timer = timeit.Timer('thing == thing2', 'from nti.schema.tests.eqhash_benchmark import Thing as Thing; thing=Thing(); thing2 = Thing()')
    times = timer.repeat()
    print("Avg Base  eq", statistics.mean(times), "stddev", statistics.stdev(times))

    timer = timeit.Timer('thing == thing2', 'from nti.schema.tests.eqhash_benchmark import ChildThing as Thing; thing=Thing(); thing2 = Thing()')
    times = timer.repeat()
    print("Avg Child eq", statistics.mean(times), "stddev", statistics.stdev(times))

    timer = timeit.Timer('thing == thing2', 'from nti.schema.tests.eqhash_benchmark import Thing2 as Thing; thing=Thing(); thing2 = Thing()')
    times = timer.repeat()
    print("Avg Super  eq", statistics.mean(times), "stddev", statistics.stdev(times))

    timer = timeit.Timer('thing == thing', 'from nti.schema.tests.eqhash_benchmark import Thing2 as Thing; thing=Thing(a={}); thing2 = Thing(a={})')
#    import cProfile
#    import pstats
#    pr = cProfile.Profile()
#    pr.enable()
    times = timer.repeat()
#    pr.disable()
#    ps = pstats.Stats(pr).sort_stats('cumulative')
#    ps.print_stats(.4)




    print("Avg Super2  eq", statistics.mean(times), "stddev", statistics.stdev(times))

    timer = timeit.Timer('thing == thing2', 'from nti.schema.tests.eqhash_benchmark import ManyThing as Thing; thing=Thing(); thing2 = Thing()')
    times = timer.repeat()
    print("Avg many  eq", statistics.mean(times), "stddev", statistics.stdev(times))


# Before
#
# Avg Base  eq 0.790581703186 stddev 0.00709198228224
# Avg Child eq 1.44241364797 stddev 0.00058100921717
# Avg Super  eq 0.772551695506 stddev 0.0120497874892
# Avg Super2  eq 0.230642795563 stddev 0.00981929676758

# Best attrgetter, params as keywords:
# Avg Base  eq 0.57781457901 stddev 0.00472933000447
# Avg Child eq 1.13719065984 stddev 0.00751860996924
# Avg Super  eq 0.576888004939 stddev 0.0073053209526
# Avg Super2  eq 0.221588929494 stddev 0.00154292380992

# Code generation
# Avg Base  eq 0.436311562856 stddev 0.0159609115497
# Avg Child eq 0.93773595492 stddev 0.0244992238919
# Avg Super  eq 0.443862199783 stddev 0.00816548353324
# Avg Super2  eq 0.216485659281 stddev 0.00651497010124

## Many attributes
# Before
# Avg Base  eq 0.762074232101 stddev 0.00893669830878
# Avg Child eq 1.46989099185 stddev 0.0260021505811
# Avg Super  eq 0.776515642802 stddev 0.011819047442
# Avg Super2  eq 0.2257057031 stddev 0.0025750486944
# Avg many  eq 1.56614136696 stddev 0.0195022584734

# Code generation
# Avg Base  eq 0.410983006159 stddev 0.00708241719015
# Avg Child eq 0.903119166692 stddev 0.0051626944104
# Avg Super  eq 0.41703470548 stddev 0.00604558003878
# Avg Super2  eq 0.208957354228 stddev 0.00508863378261
# Avg many  eq 0.797417243322 stddev 0.0198358058579

if __name__ == '__main__':
    import sys
    if '--timehash' in sys.argv:
        bench_hash()
    elif '--timeeq' in sys.argv:
        bench_eq()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for jsonschema.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import unittest

from zope.interface import Interface
from zope.interface import Attribute

from .. import jsonschema
from ..field import DecodingValidTextLine
from ..field import ListOrTuple
from ..field import Dict
from ..field import List
from ..field import Real
from ..field import Rational
from ..field import Complex
from ..field import Integral

# ABCs
from ..field import Sequence
from ..field import IndexedIterable
from ..field import MutableSequence

from ..field import Mapping
from ..field import MutableMapping


from hamcrest import assert_that
from hamcrest import has_entry
from hamcrest import is_
from hamcrest import has_length
from hamcrest import has_key
from hamcrest import is_not
from hamcrest import same_instance
from hamcrest import none
does_not = is_not


__docformat__ = "restructuredtext en"


#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904,inherit-non-class,no-method-argument

class TranslateTestSchema(jsonschema.JsonSchemafier):

    def _translate(self, text):
        return text + self.context


class TestJsonSchemafier(unittest.TestCase):

    def test_application_info(self):

        app_info = {
            'text_key': u'a value',
            'byte_key': b'bytes',
            'int_key': 7,
            'list_key': [42,]
        }

        class IA(Interface):

            field = Attribute("A field")
            field.setTaggedValue(jsonschema.TAG_APPLICATION_INFO, app_info)

        schema = TranslateTestSchema(IA, context=u' TEST').make_schema()

        assert_that(schema, has_key("field"))
        field = schema['field']
        assert_that(field, has_key("application_info"))
        # The dict itself was copied
        assert_that(field['application_info'], is_not(same_instance(app_info)))
        # as were its contents
        assert_that(field['application_info'], has_entry('list_key', is_([42])))
        assert_that(field['application_info'],
                    has_entry('list_key', is_not(same_instance(app_info['list_key']))))

        # Text strings were translated
        assert_that(field['application_info'], has_entry('text_key', u'a value TEST'))

        # Byte strings were not (is that even legel in json?)
        assert_that(field['application_info'], has_entry('byte_key', b'bytes'))

    def test_excludes(self):

        class IA(Interface):

            def method():
                "A method"

            _thing = Attribute("A private attribute")

            hidden = Attribute("Hidden attribute")
            hidden.setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)

        schema = jsonschema.JsonSchemafier(IA).make_schema()
        assert_that(schema, is_({}))

    def test_readonly_override(self):
        class IA(Interface):

            field = Attribute("A field")

        schema = jsonschema.JsonSchemafier(IA, readonly_override=True).make_schema()
        assert_that(schema, has_entry('field', has_entry('readonly', True)))

    def test_ui_type(self):
        class IA(Interface):

            field = Attribute("A field")
            field.setTaggedValue(jsonschema.TAG_UI_TYPE, 'MyType')

        schema = jsonschema.JsonSchemafier(IA).make_schema()
        assert_that(schema, has_entry('field', has_entry('type', 'MyType')))

    def test_type_from_types(self):
        from zope.schema import Object
        from nti.schema.field import Variant

        def _assert_type(t, name='field',
                         **kwargs):
            schema = TranslateTestSchema(IA, context=' TEST').make_schema()
            assert_that(schema, has_entry(name, has_entry('type', t)))
            for nested_name, matcher in kwargs.items():
                assert_that(schema, has_entry(name, has_entry(nested_name, matcher)))

            return schema

        class IUnderlyingA(Interface):

            text_field = DecodingValidTextLine(title=u'nested text',
                                               min_length=5)

            hidden_text_field = DecodingValidTextLine()

        class IUnderlyingB(Interface):

            int_field = Integral()

        class IA(Interface):

            field = Attribute("A field")

            field2 = DecodingValidTextLine(title=u'A title',
                                           description=u'A description')

            list_field = List(DecodingValidTextLine())
            list_or_tuple_field = ListOrTuple(Real())
            dict_field = Dict(DecodingValidTextLine(), Real())
            mapping_field = Mapping()
            mmapping_field = MutableMapping()
            sequence_field = Sequence()
            msequence_field = MutableSequence()
            iiterable_field = IndexedIterable()

            real_field = Real()
            rational_field = Rational()
            complex_field = Complex()
            integral_field = Integral()

            object_field = Object(IUnderlyingA, required=True)
            variant_field = Variant([Object(IUnderlyingA),
                                     Object(IUnderlyingB),
                                     DecodingValidTextLine()])

            hidden_object_field = Object(IUnderlyingA, required=True)

        IA['field']._type = str
        _assert_type('string')

        IA['field']._type = (str,)
        _assert_type('string')


        IA['field']._type = float
        _assert_type('float')

        IA['field']._type = (float,)
        _assert_type('float')

        IA['field']._type = int
        _assert_type('int')

        IA['field']._type = (int,)
        _assert_type('int')

        IA['field']._type = bool
        _assert_type('bool')

        _assert_type('string', 'field2',
                     base_type='string',
                     title=u'A title TEST',
                     description=u'A description TEST')

        _assert_type('list', 'list_field', value_type=has_entry('type', 'string'))
        _assert_type('list', 'list_or_tuple_field')
        _assert_type('list', 'sequence_field')
        _assert_type('list', 'msequence_field',
                     unique=False, value_type=none(), min_length=0, max_length=none())
        _assert_type('list', 'iiterable_field')

        _assert_type('dict', 'dict_field',
                     key_type=has_entry('type', 'string'),
                     value_type=has_entry('type', 'Real'))
        _assert_type('dict', 'mapping_field')
        _assert_type('dict', 'mmapping_field')

        _assert_type('Real', 'real_field', base_type='float')
        _assert_type('Rational', 'rational_field',
                     base_type='float', max=none(), min=none())
        _assert_type('Complex', 'complex_field', base_type='float')
        _assert_type('Integral', 'integral_field', base_type='int')

        _assert_type('Object', 'object_field',
                     schema=has_entry('text_field', has_entry('min_length', 5)))

        # Now hiding some fields
        # Top-level fields can be hidden
        IA['hidden_object_field'].setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)

        # Fields of a schema nested in an object can be hidden.
        IUnderlyingA['hidden_text_field'].setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)

        # A field of a variant can be hidden
        IA['variant_field'].fields[2].setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)
        # All fields of lists and dicts can be hidden.
        IA['dict_field'].key_type.setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)
        IA['dict_field'].value_type.setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)
        IA['list_field'].value_type.setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)


        schema = _assert_type('Variant', 'variant_field',
                              value_type_options=has_length(2))
        variant_schema_values = schema['variant_field']['value_type_options']
        assert_that(variant_schema_values[0], has_entry('type', 'Object'))

        assert_that(schema, does_not(has_key('hidden_object_field')))
        assert_that(schema['object_field']['schema'], does_not(has_key('hidden_text_field')))

        assert_that(schema['list_field'], has_entry('value_type', None))
        assert_that(schema['dict_field'], has_entry('value_type', None))
        assert_that(schema['dict_field'], has_entry('key_type', None))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from six import text_type
from zope.interface import Interface
from zope.interface import classImplements
from zope.testing import cleanup

from nti.testing.layers import ConfiguringLayerMixin
from nti.testing.layers import ZopeComponentLayer

__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904,inherit-non-class

class SchemaLayer(ZopeComponentLayer,
                  ConfiguringLayerMixin):

    set_up_packages = ('nti.schema',)

    @classmethod
    def setUp(cls):
        cls.setUpPackages()

    @classmethod
    def tearDown(cls):
        cls.tearDownPackages()
        cleanup.cleanUp()

    @classmethod
    def testSetUp(cls):
        pass

    testTearDown = testSetUp


class IUnicode(Interface):
    "Unicode strings"


classImplements(text_type, IUnicode)



#!/usr/bin/env python
import codecs
from setuptools import setup, find_packages

version = '1.12.0'

entry_points = {
}


def _read(fname):
    with codecs.open(fname, encoding='utf-8') as f:
        return f.read()


setup(
    name='nti.schema',
    version=version,
    author='Jason Madden',
    author_email='open-source@nextthought.com',
    description=('Zope schema related support'),
    long_description=(
        _read('README.rst')
        + '\n\n'
        + _read('CHANGES.rst')
    ),
    license='Apache',
    keywords='zope schema',
    url='https://github.com/NextThought/nti.schema',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Framework :: Zope3',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'Acquisition',
        'nti.i18n',
        'six',
        'setuptools',
        'zope.event',
        'zope.schema >= 4.8.0',
        'zope.i18n',
        'zope.i18nmessageid',
        'zope.vocabularyregistry',
        'zope.deprecation',
        'zope.deferredimport >= 4.2.1',
    ],
    extras_require={
        'test': [
            'pyhamcrest',
            'nti.testing',
            'zope.testrunner',
        ],
        'docs': [
            'Sphinx',
            'nti.testing',
            'repoze.sphinx.autointerface',
            'sphinx_rtd_theme',
        ],
    },
    namespace_packages=['nti'],
    entry_points=entry_points,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Computed attributes based on schema fields.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# stdlib imports
import sys

from Acquisition import aq_base
from Acquisition.interfaces import IAcquirer
from zope.schema import interfaces as sch_interfaces
from zope.schema.fieldproperty import FieldProperty
from zope.schema.fieldproperty import FieldPropertyStoredThroughField
from zope.schema.fieldproperty import createFieldProperties

__docformat__ = "restructuredtext en"

class AcquisitionFieldProperty(FieldProperty):
    """
    A field property that supports acquisition. Returned objects
    will be ``__of__`` the instance, and set objects will always be the unwrapped
    base.
    """

    def __get__(self, instance, klass):
        result = super(AcquisitionFieldProperty, self).__get__(instance, klass)
        if instance is not None and IAcquirer.providedBy(result):  # even defaults get wrapped
            result = result.__of__(instance)
        return result

    def __set__(self, instance, value):
        super(AcquisitionFieldProperty, self).__set__(instance, aq_base(value))

class UnicodeConvertingFieldProperty(FieldProperty):
    """
    Accepts bytes input for the unicode property if it can be
    decoded as UTF-8. This is primarily to support legacy test cases
    and should be removed when all constants are unicode.
    """

    def __set__(self, inst, value):
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        super(UnicodeConvertingFieldProperty, self).__set__(inst, value)

def _find_schema_from_field(field):
    if not sch_interfaces.IObject.providedBy(field) and not hasattr(field, 'schema'):
        raise sch_interfaces.WrongType("Don't know how to get schema from %s" % field)
    return field.schema

def _make_adapter_set(klass):

    def __set__(self, inst, value):
        try:
            super(klass, self).__set__(inst, value)
        except sch_interfaces.SchemaNotProvided:
            try:
                value = self.schema(value)
            except TypeError:
                # Let's raise the better error when we call again
                pass

            super(klass, self).__set__(inst, value)

    return __set__

class AdaptingFieldProperty(FieldProperty):
    """
    Primarily for legacy support and testing, adds adaptation to an interface
    when setting a field. This is most useful when the values are simple literals
    like strings.
    """

    def __init__(self, field, name=None):
        self.schema = _find_schema_from_field(field)
        super(AdaptingFieldProperty, self).__init__(field, name=name)

AdaptingFieldProperty.__set__ = _make_adapter_set(AdaptingFieldProperty)

class AdaptingFieldPropertyStoredThroughField(FieldPropertyStoredThroughField):
    """
    Primarily for legacy support and testing, adds adaptation to an interface
    when setting a field. This is most useful when the values are simple literals
    like strings.
    """

    def __init__(self, field, name=None):
        self.schema = _find_schema_from_field(field)
        super(AdaptingFieldPropertyStoredThroughField, self).__init__(field, name=name)

AdaptingFieldPropertyStoredThroughField.__set__ = _make_adapter_set(
    AdaptingFieldPropertyStoredThroughField)

def createDirectFieldProperties(__schema, omit=(), adapting=False):
    """
    Like :func:`zope.schema.fieldproperty.createFieldProperties`, except
    only creates properties for fields directly contained within the
    given schema; inherited fields from parent interfaces are assummed
    to be implemented in a base class of the current class::

      >>> from zope import interface
      >>> from nti.schema.field import TextLine, Object
      >>> class IA(interface.Interface):
      ...    a = TextLine(title=u"a")

      >>> class IB(IA):
      ...    b = Object(interface.Interface)

      >>> class A(object):
      ...    createFieldProperties(IA)

      >>> class B(object):
      ...    createDirectFieldProperties(IB, adapting=True)

      >>> 'a' in A.__dict__
      True
      >>> 'a' in B.__dict__
      False
      >>> 'b' in B.__dict__
      True


    :keyword adapting: If set to ``True`` (not the default), fields
        that implement :class:`.IObject` will use an :class:`AdaptingFieldProperty`.
    """

    __my_names = set(__schema.names())
    __all_names = set(__schema.names(all=True))

    __not_my_names = __all_names - __my_names
    __not_my_names.update(omit)

    # The existing implementation relies on getframe(1) to find the caller,
    # which is us. So we do the same and copy to the real caller
    __frame = None
    __before = None
    __before = list(locals().keys())
    createFieldProperties(__schema, omit=__not_my_names)

    __frame = sys._getframe(1) # pylint:disable=protected-access
    for k, v in list(locals().items()):
        if k not in __before:
            if adapting and sch_interfaces.IObject.providedBy(__schema[k]):
                v = AdaptingFieldProperty(__schema[k])
            __frame.f_locals[k] = v


def field_name(field):
    """
    Produce a clean version of a field's name.

    The
    :class:`zope.schema.fieldproperty.FieldPropertyStoredThroughField`
    class mangles the field name, making it difficult to trace fields
    back to their intended attribute name. This undoes that mangling
    if possible.

    The field in an interface has a normal name::

        >>> from zope.schema.fieldproperty import FieldPropertyStoredThroughField
        >>> from zope.schema import Field
        >>> from zope import interface
        >>> class IA(interface.Interface):
        ...    a = Field()
        >>> IA['a'].__name__
        'a'

    The field as stored by a ``FieldProperty`` has a normal name::

        >>> from zope.schema.fieldproperty import FieldProperty
        >>> class A(object):
        ...    createFieldProperties(IA)
        >>> A.a.__name__
        'a'

    But using a ``FieldPropertyStoredThroughField`` mangles the name
    (whether accessed through the ``FieldPropertyStoredThroughField``
    or directly)::

        >>> from zope.schema.fieldproperty import FieldPropertyStoredThroughField
        >>> class A2(object):
        ...    a = FieldPropertyStoredThroughField(IA['a'])
        >>> A2.a.__name__
        '__st_a_st'
        >>> A2.a.field.__name__
        '__st_a_st'

    This function demangles the name (whether accessed through the
    ``FieldPropertyStoredThroughField`` or directly)::

        >>> from nti.schema.fieldproperty import field_name
        >>> field_name(A2.a)
        'a'
        >>> field_name(A2.a.field)
        'a'

    Without damaging the names of regular fields or regular
    ``FieldProperty`` fields::

        >>> field_name(IA['a'])
        'a'
        >>> field_name(A.a)
        'a'

    .. versionadded:: 1.10.0
    """
    if field.__name__ and field.__name__.startswith('__st_') and field.__name__.endswith('_st'):
        return field.__name__[5:-3]
    return field.__name__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Vocabularies and factories for use in schema fields.

When this package is configured (via ``configure.zcml``) there will be a schema
vocabulary named ``Countries`` available::

  >>> from nti.schema.field import Choice
  >>> from zope import interface

  >>> class IA(interface.Interface):
  ...      choice = Choice(title=u"Choice",
  ...                      vocabulary="Countries")
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope import component
from zope.schema.vocabulary import SimpleTerm as _SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary as _SimpleVocabulary

from nti.i18n.locales.interfaces import ICountryAvailability

__docformat__ = "restructuredtext en"


class CountryTerm(_SimpleTerm):
    """
    A titled, tokenized term representing a country. The
    token is the ISO3166 country code. The ``flag`` value is a
    browserresource path to an icon representing the country.
    """

    def __init__(self, *args, **kwargs):
        self.flag = kwargs.pop('flag', None)
        super(CountryTerm, self).__init__(*args, **kwargs)

    @classmethod
    def fromItem(cls, item):
        token, cdata = item
        value = cdata['name']
        title = value
        flag = cdata['flag']

        return cls(value, token, title, flag=flag)

    def toExternalObject(self):
        return {
            'token': self.token,
            'title': self.title,
            'value': self.value,
            'flag': self.flag
        }

class _CountryVocabulary(_SimpleVocabulary):
    """
    ``__contains__`` is based on the token, not the value.
    """

    def __contains__(self, token):
        return token in self.by_token

def CountryVocabularyFactory(context):
    """
    A vocabulary factory.
    """
    countries = component.getUtility(ICountryAvailability)
    return _CountryVocabulary([CountryTerm.fromItem(item)
                               for item
                               in countries.getCountries().items()])
